import logging

from ..helpers import LoggedFunction
from ..settings import config, current_changelog_components

from .changelog import changelog_headers, changelog_table  # noqa isort:skip
from .compare import compare_url  # noqa isort:skip

logger = logging.getLogger(__name__)


@LoggedFunction(logger)
def markdown_changelog(
    owner: str,
    repo_name: str,
    version: str,
    changelog: dict,
    header: bool = False,
    previous_version: str = None,
) -> str:
    """
    Generate a markdown version of the changelog.

    :param owner: The repo owner.
    :param repo_name: The repo name.
    :param version: A string with the version number.
    :param previous_version: A string with the last version number, to
        use for the comparison URL. If omitted, the URL will not be included.
    :param changelog: A parsed changelog dict from generate_changelog.
    :param header: A boolean that decides whether a version number header should be included.
    :return: The markdown formatted changelog.
    """
    output = f"## v{version}\n" if header else ""

    # Add the output of each component separated by a blank line
    output += "\n\n".join(
        (
            component_output.strip()
            for component_output in (
                component(
                    owner=owner,
                    repo_name=repo_name,
                    version=version,
                    previous_version=previous_version,
                    changelog=changelog,
                    changelog_sections=config.get("changelog_sections").split(","),
                )
                for component in current_changelog_components()
            )
            if component_output is not None
        )
    )

    return output



"""Commit parser which looks for emojis to determine the type of commit"""
import logging
import re
from typing import Optional

from ..helpers import LoggedFunction
from ..settings import config
from .parser_helpers import ParsedCommit, parse_paragraphs

logger = logging.getLogger(__name__)


@LoggedFunction(logger)
def parse_commit_message(
    message: str,
) -> ParsedCommit:
    """
    Parse a commit using an emoji in the subject line.

    When multiple emojis are encountered, the one with the highest bump
    level is used. If there are multiple emojis on the same level, the
    we use the one listed earliest in the configuration.

    If the message does not contain any known emojis, then the level to bump
    will be 0 and the type of change "Other". This parser never raises
    UnknownCommitMessageStyleError.

    Emojis are not removed from the description, and will appear alongside
    the commit subject in the changelog.

    :param message: A string of a commit message.
    :return: A tuple of (level to bump, type of change, scope of change, a tuple with descriptions)
    """

    subject = message.split("\n")[0]

    major = config.get("major_emoji").split(",")
    minor = config.get("minor_emoji").split(",")
    patch = config.get("patch_emoji").split(",")
    all_emojis = major + minor + patch

    # Loop over emojis from most important to least important
    # Therefore, we find the highest level emoji first
    primary_emoji = "Other"
    for emoji in all_emojis:
        if emoji in subject:
            primary_emoji = emoji
            break
    logger.debug(f"Selected {primary_emoji} as the primary emoji")

    # Find which level this commit was from
    level_bump = 0
    if primary_emoji in major:
        level_bump = 3
    elif primary_emoji in minor:
        level_bump = 2
    elif primary_emoji in patch:
        level_bump = 1

    # All emojis will remain part of the returned description
    descriptions = parse_paragraphs(message)
    return ParsedCommit(
        level_bump,
        primary_emoji,
        None,
        descriptions,
        descriptions[1:] if level_bump == 3 else [],
    )



"""History
"""
import csv
import logging
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional, Set, Union

import semver
import tomlkit
from dotty_dict import Dotty

from ..errors import ImproperConfigurationError
from ..helpers import LoggedFunction
from ..settings import config
from ..vcs_helpers import get_commit_log, get_formatted_tag, get_last_version
from .logs import evaluate_version_bump  # noqa

from .parser_angular import parse_commit_message as angular_parser  # noqa isort:skip
from .parser_emoji import parse_commit_message as emoji_parser  # noqa isort:skip
from .parser_scipy import parse_commit_message as scipy_parser  # noqa isort:skip
from .parser_tag import parse_commit_message as tag_parser  # noqa isort:skip

logger = logging.getLogger(__name__)


class VersionDeclaration(ABC):
    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    @staticmethod
    def from_toml(config_str: str):
        """
        Instantiate a `TomlVersionDeclaration` from a string specifying a path and a key
        matching the version number.
        """
        path, key = config_str.split(":", 1)
        return TomlVersionDeclaration(path, key)

    @staticmethod
    def from_variable(config_str: str):
        """
        Instantiate a `PatternVersionDeclaration` from a string specifying a path and a
        variable name.
        """
        path, variable = config_str.split(":", 1)
        pattern = (
            rf'{variable} *[:=] *["\']{PatternVersionDeclaration.version_regex}["\']'
        )
        return PatternVersionDeclaration(path, pattern)

    @staticmethod
    def from_pattern(config_str: str):
        """
        Instantiate a `PatternVersionDeclaration` from a string specifying a path and a
        regular expression matching the version number.
        """
        path, pattern = config_str.split(":", 1)
        pattern = pattern.format(version=PatternVersionDeclaration.version_regex)
        return PatternVersionDeclaration(path, pattern)

    @abstractmethod
    def parse(self) -> Set[str]:
        """
        Return the versions.

        Because a source can match in multiple places, this method returns a
        set of matches. Generally, there should only be one element in this
        set (i.e. even if the version is specified in multiple places, it
        should be the same version in each place), but it falls on the caller
        to check for this condition.
        """

    @abstractmethod
    def replace(self, new_version: str):
        """
        Update the versions.

        This method reads the underlying file, replaces each occurrence of the
        matched pattern, then writes the updated file.

        :param new_version: The new version number as a string
        """


class TomlVersionDeclaration(VersionDeclaration):
    def __init__(self, path, key):
        super().__init__(path)
        self.key = key

    def _read(self) -> Dotty:
        toml_doc = tomlkit.loads(self.path.read_text())
        return Dotty(toml_doc)

    def parse(self) -> Set[str]:
        _config = self._read()
        if self.key in _config:
            return {_config.get(self.key)}
        return set()

    def replace(self, new_version: str) -> None:
        _config = self._read()
        if self.key in _config:
            _config[self.key] = new_version
            self.path.write_text(tomlkit.dumps(_config))


class PatternVersionDeclaration(VersionDeclaration):
    """
    Represent a version number in a particular file.

    The version number is identified by a regular expression.  Methods are
    provided both the read the version number from the file, and to update the
    file with a new version number.  Use the `load_version_patterns()` factory
    function to create the version patterns specified in the config files.
    """

    version_regex = r"(\d+\.\d+(?:\.\d+)?)"

    # The pattern should be a regular expression with a single group,
    # containing the version to replace.
    def __init__(self, path: str, pattern: str):
        super().__init__(path)
        self.pattern = pattern

    def parse(self) -> Set[str]:
        """
        Return the versions matching this pattern.

        Because a pattern can match in multiple places, this method returns a
        set of matches.  Generally, there should only be one element in this
        set (i.e. even if the version is specified in multiple places, it
        should be the same version in each place), but it falls on the caller
        to check for this condition.
        """
        content = self.path.read_text()

        versions = {
            m.group(1) for m in re.finditer(self.pattern, content, re.MULTILINE)
        }

        logger.debug(
            f"Parsing current version: path={self.path!r} pattern={self.pattern!r} num_matches={len(versions)}"
        )
        return versions

    def replace(self, new_version: str):
        """
        Update the versions matching this pattern.

        This method reads the underlying file, replaces each occurrence of the
        matched pattern, then writes the updated file.

        :param new_version: The new version number as a string
        """
        n = 0
        old_content = self.path.read_text()

        def swap_version(m):
            nonlocal n
            n += 1
            s = m.string
            i, j = m.span()
            ii, jj = m.span(1)
            return s[i:ii] + new_version + s[jj:j]

        new_content = re.sub(
            self.pattern, swap_version, old_content, flags=re.MULTILINE
        )

        logger.debug(
            f"Writing new version number: path={self.path!r} pattern={self.pattern!r} num_matches={n!r}"
        )

        self.path.write_text(new_content)


def get_prerelease_pattern() -> str:
    return "-" + config.get("prerelease_tag") + "."


@LoggedFunction(logger)
def get_current_version_by_tag(omit_pattern=None) -> str:
    """
    Find the current version of the package in the current working directory using git tags.

    :return: A string with the version number or 0.0.0 on failure.
    """
    version = get_last_version(omit_pattern=omit_pattern)
    if version:
        return version

    logger.debug("no version found, returning default of v0.0.0")
    return "0.0.0"


@LoggedFunction(logger)
def get_current_version_by_config_file(omit_pattern=None) -> str:
    """
    Get current version from the version variable defined in the configuration.

    :return: A string with the current version number
    :raises ImproperConfigurationError: if either no versions are found, or
    multiple versions are found.
    """
    declarations = load_version_declarations()
    versions = set.union(*[x.parse() for x in declarations])

    if len(versions) == 0:
        raise ImproperConfigurationError(
            "no versions found in the configured locations"
        )
    if len(versions) != 1:
        version_strs = ", ".join(repr(x) for x in versions)
        raise ImproperConfigurationError(f"found conflicting versions: {version_strs}")

    version = versions.pop()
    logger.debug(f"Regex matched version: {version}")
    return version


def get_current_version(prerelease_version: bool = False) -> str:
    """
    Get current version from tag or version variable, depending on configuration.
    This will not return prerelease versions.

    :return: A string with the current version number
    """
    omit_pattern = None if prerelease_version else get_prerelease_pattern()
    if config.get("version_source") in ["tag", "tag_only"]:
        return get_current_version_by_tag(omit_pattern)
    current_version = get_current_version_by_config_file(omit_pattern)
    if omit_pattern and omit_pattern in current_version:
        return get_previous_version(current_version)
    return current_version


@LoggedFunction(logger)
def get_new_version(
    current_version: str, level_bump: str, prerelease: bool = False
) -> str:
    """
    Calculate the next version based on the given bump level with semver.

    :param current_version: The version the package has now.
    :param level_bump: The level of the version number that should be bumped.
        Should be `'major'`, `'minor'` or `'patch'`.
    :param prerelease: Should the version bump be marked as a prerelease
    :return: A string with the next version number.
    """
    if not level_bump:
        logger.debug("No bump requested, using input version")
        new_version = current_version
    else:
        new_version = str(
            semver.VersionInfo.parse(current_version).next_version(part=level_bump)
        )

    if prerelease:
        logger.debug("Prerelease requested")
        potentialy_prereleased_current_version = get_current_version(
            prerelease_version=True
        )
        if get_prerelease_pattern() in potentialy_prereleased_current_version:
            logger.debug("Previouse prerelease detected, increment prerelease version")
            prerelease_num = (
                int(potentialy_prereleased_current_version.split(".")[-1]) + 1
            )
        else:
            logger.debug("No previouse prerelease detected, starting from 0")
            prerelease_num = 0
        new_version = new_version + get_prerelease_pattern() + str(prerelease_num)

    return new_version


@LoggedFunction(logger)
def get_previous_version(version: str, omit_pattern: str = None) -> Optional[str]:
    """
    Return the version prior to the given version.

    :param version: A string with the version number.
    :return: A string with the previous version number.
    """
    found_version = False
    for commit_hash, commit_message in get_commit_log():
        logger.debug(f"Checking commit {commit_hash}")
        if version in commit_message:
            found_version = True
            logger.debug(f'Found version in commit "{commit_message}"')
            continue

        if found_version:
            if omit_pattern and omit_pattern in commit_message:
                continue
            matches = re.match(r"v?(\d+.\d+.\d+)", commit_message)
            if matches:
                logger.debug(f"Version matches regex {commit_message}")
                return matches.group(1).strip()

    return get_last_version(
        [version, get_formatted_tag(version)], omit_pattern=omit_pattern
    )


@LoggedFunction(logger)
def set_new_version(new_version: str) -> bool:
    """
    Update the version number in each configured location.

    :param new_version: The new version number as a string.
    :return: `True` if it succeeded.
    """

    for declaration in load_version_declarations():
        declaration.replace(new_version)

    return True


def load_version_declarations() -> List[VersionDeclaration]:
    """
    Create the `VersionDeclaration` objects specified by the config file.
    """
    declarations = []

    def iter_fields(x):
        if not x:
            return
        if isinstance(x, list):
            yield from x
        else:
            # Split by commas, but allow the user to escape commas if
            # necessary.
            yield from next(csv.reader([x]))

    for version_var in iter_fields(config.get("version_variable")):
        declaration = VersionDeclaration.from_variable(version_var)
        declarations.append(declaration)
    for version_pat in iter_fields(config.get("version_pattern")):
        declaration = VersionDeclaration.from_pattern(version_pat)
        declarations.append(declaration)
    for version_toml in iter_fields(config.get("version_toml")):
        declaration = VersionDeclaration.from_toml(version_toml)
        declarations.append(declaration)

    if not declarations:
        raise ImproperConfigurationError(
            "must specify either 'version_variable', 'version_pattern' or 'version_toml'"
        )

    return declarations



from semantic_release.history.parser_helpers import parse_paragraphs


def test_parse_paragraphs_with_no_content():
    paragraphs = parse_paragraphs("")
    assert len(paragraphs) == 0


def test_parse_paragraphs():
    paragraphs = parse_paragraphs("Paragraph 0\n\nParagraph 1")
    assert len(paragraphs) == 2
    assert paragraphs[0] == "Paragraph 0"
    assert paragraphs[1] == "Paragraph 1"


def test_parse_paragraphs_multiline():
    paragraphs = parse_paragraphs("This paragraph is over\nmultiple lines.")
    assert paragraphs[0] == "This paragraph is over multiple lines."



"""Helper for using Twine to upload to an artifact repository.
"""
import logging
import os
from dataclasses import InitVar
from dataclasses import asdict as dataclass_asdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional

import requests
from twine.commands.upload import upload as twine_upload
from twine.exceptions import TwineException
from twine.settings import Settings as TwineSettings

from semantic_release import ImproperConfigurationError
from semantic_release.helpers import LoggedFunction
from semantic_release.settings import config

logger = logging.getLogger(__name__)


def get_env_var(name: str) -> Optional[str]:
    """
    Resolve variable name from config and return matching environment variable

    :param name: Variable name to retrieve from environment

    :returns Value of environment variable or None if not set.
    """
    return os.environ.get(config.get(name))


@dataclass(eq=False)
class ArtifactRepo:
    """
    Object that manages the configuration and execution of upload using Twine.

    This object needs only one shared argument to be instantiated.
    """

    dist_path: InitVar[Path]
    repository_name: str = "pypi"
    repository_url: Optional[str] = None
    username: Optional[str] = field(repr=False, default=None)
    password: Optional[str] = field(repr=False, default=None)
    dists: List[str] = field(init=False, default_factory=list)

    def __post_init__(self, dist_path: Path) -> None:
        """
        :param dist_path: Path to dist folder containing the files to upload.
        """
        self._handle_credentials_init()
        self._handle_repository_config()
        self._handle_glob_patterns(dist_path)

    @LoggedFunction(logger)
    def _handle_credentials_init(self) -> None:
        """
        Initialize credentials from environment variables.

        For the transitional period until the *pypi* variables can be safely removed,
        additional complexity is needed.

        :raises ImproperConfigurationError:
            Error while setting up credentials configuration.
        """
        username = get_env_var("repository_user_var") or get_env_var("pypi_user_var")
        password = (
            get_env_var("repository_pass_var")
            or get_env_var("pypi_pass_var")
            or get_env_var("pypi_token_var")
        )
        if username and password:
            self.username = username
            self.password = password
        elif password and not username:
            self.username = username or "__token__"
            self.password = password
            logger.warning(
                "Providing only password or token without username is deprecated"
            )
        # neither username nor password provided, check for ~/.pypirc file
        elif not Path("~/.pypirc").expanduser().exists():
            raise ImproperConfigurationError(
                "Missing credentials for uploading to artifact repository"
            )

    @LoggedFunction(logger)
    def _handle_glob_patterns(self, dist_path: Path) -> None:
        """
        Load glob patterns that select the distribution files to publish.

        :param dist_path: Path to folder with package files
        """
        glob_patterns = config.get("dist_glob_patterns") or config.get(
            "upload_to_pypi_glob_patterns"
        )
        glob_patterns = (glob_patterns or "*").split(",")

        self.dists = [str(dist_path.joinpath(pattern)) for pattern in glob_patterns]

    @LoggedFunction(logger)
    def _handle_repository_config(self) -> None:
        """
        Initialize repository settings from config.

        *repository_url* overrides *repository*, Twine handles this the same way.

        Defaults to repository_name `pypi` when both are not set.
        """
        repository_url = config.get("repository_url")
        repository_name = config.get("repository")

        if repository_url:
            self.repository_url = repository_url
        elif repository_name:
            self.repository_name = repository_name

    @LoggedFunction(logger)
    def _create_twine_settings(self, addon_kwargs: Dict[str, Any]) -> TwineSettings:
        """
        Gather all parameters that had a value set during instantiation and
        pass them to Twine which then validates and laods the config.
        """
        params = {name: val for name, val in dataclass_asdict(self).items() if val}
        settings = TwineSettings(**params, **addon_kwargs)

        return settings

    @LoggedFunction(logger)
    def upload(
        self, noop: bool, verbose: bool, skip_existing: bool, **additional_kwargs
    ) -> bool:
        """
        Upload artifact to repository using Twine.

        For known repositories (like PyPI), the web URLs of successfully uploaded packages
        will be displayed.

        :param noop: Do not apply any changes..
        :param verbose: Show verbose output for Twine.
        :param skip_existing: Continue uploading files if one already exists.
            (May not work, check your repository for support.)

        :raises ImproperConfigurationError:
            The upload failed due to a configuration error.

        :returns True if successfull, False otherwise.
        """
        addon_kwargs = {
            "non_interactive": True,
            "verbose": verbose,
            "skip_existing": skip_existing,
            **additional_kwargs,
        }

        try:
            twine_settings = self._create_twine_settings(addon_kwargs)
            if not noop:
                twine_upload(upload_settings=twine_settings, dists=self.dists)
        except TwineException as e:
            raise ImproperConfigurationError(
                "Upload to artifact repository has failed"
            ) from e
        except requests.HTTPError as e:
            logger.warning(f"Upload to artifact repository has failed: {e}")
            return False
        else:
            return True

    @staticmethod
    def upload_enabled() -> bool:
        """
        Check if artifact repository upload is enabled

        :returns True if upload is enabled, False otherwise.
        """
        return config.get("upload_to_repository") and config.get("upload_to_pypi")



MAJOR = (
    "221",
    "feat(x): Add super-feature\n\n"
    "BREAKING CHANGE: Uses super-feature as default instead of dull-feature.",
)
MAJOR2 = (
    "222",
    "feat(x): Add super-feature\n\nSome explanation\n\n"
    "BREAKING CHANGE: Uses super-feature as default instead of dull-feature.",
)
MAJOR_MENTIONING_1_0_0 = (
    "223",
    "feat(x): Add super-feature\n\nSome explanation\n\n"
    "BREAKING CHANGE: Uses super-feature as default instead of dull-feature from v1.0.0.",
)
MAJOR_MULTIPLE_FOOTERS = (
    "244",
    "feat(x): Lots of breaking changes\n\n"
    "BREAKING CHANGE: Breaking change 1\n\n"
    "Not a BREAKING CHANGE\n\n"
    "BREAKING CHANGE: Breaking change 2",
)
MAJOR_EXCL_WITH_FOOTER = (
    "231",
    "feat(x)!: Add another feature\n\n"
    "BREAKING CHANGE: Another feature, another breaking change",
)
MAJOR_EXCL_NOT_FOOTER = (
    "232",
    "fix!: Fix a big bug that everyone exploited\n\nThis is the reason you should not exploit bugs",
)
MINOR = ("111", "feat(x): Add non-breaking super-feature")
PATCH = ("24", "fix(x): Fix bug in super-feature")
NO_TAG = ("191", "docs(x): Add documentation for super-feature")
UNKNOWN_STYLE = ("7", "random commits are the worst")

ALL_KINDS_OF_COMMIT_MESSAGES = [MINOR, MAJOR, MINOR, PATCH]
MINOR_AND_PATCH_COMMIT_MESSAGES = [MINOR, PATCH]
PATCH_COMMIT_MESSAGES = [PATCH, PATCH]
MAJOR_LAST_RELEASE_MINOR_AFTER = [MINOR, ("22", "1.1.0"), MAJOR]
MAJOR_MENTIONING_LAST_VERSION = [MAJOR_MENTIONING_1_0_0, ("22", "1.0.0"), MAJOR]



from pathlib import Path
from textwrap import dedent

import mock
import pytest

import semantic_release
from semantic_release.history import (
    ImproperConfigurationError,
    PatternVersionDeclaration,
    TomlVersionDeclaration,
    VersionDeclaration,
    get_current_version,
    get_current_version_by_config_file,
    get_current_version_by_tag,
    get_new_version,
    get_previous_version,
    load_version_declarations,
    set_new_version,
)

from .. import wrapped_config_get


@pytest.fixture
def tmp_cwd(tmp_path):
    import os

    try:
        orig_path = os.getcwd()
        os.chdir(tmp_path)
        yield tmp_path
    finally:
        os.chdir(orig_path)


def test_current_version_should_return_correct_version():
    assert get_current_version() == semantic_release.__version__


@mock.patch("semantic_release.history.get_last_version", return_value="last_version")
def test_current_version_should_return_git_version(mock_last_version):
    assert "last_version" == get_current_version_by_tag()


@mock.patch(
    "semantic_release.history.config.get", wrapped_config_get(version_source="tag")
)
@mock.patch("semantic_release.history.get_last_version", return_value=None)
def test_current_version_should_return_default_version(mock_last_version):
    assert "0.0.0" == get_current_version()


@mock.patch(
    "semantic_release.history.config.get", wrapped_config_get(version_source="tag_only")
)
def test_current_version_should_run_with_tag_only(mocker):
    mock_get_current_version_by_tag = mocker.patch(
        "semantic_release.history.get_current_version_by_tag", return_value=None
    )
    mock_get_current_version_by_config_file = mocker.patch(
        "semantic_release.history.get_current_version_by_config_file", return_value=None
    )

    get_current_version()

    assert mock_get_current_version_by_tag.called
    assert not mock_get_current_version_by_config_file.called


class TestGetPreviousVersion:
    @mock.patch(
        "semantic_release.history.get_commit_log",
        lambda: [("211", "0.10.0"), ("13", "0.9.0")],
    )
    def test_should_return_correct_version(self):
        assert get_previous_version("0.10.0") == "0.9.0"

    @mock.patch(
        "semantic_release.history.get_commit_log",
        lambda: [("211", "0.10.0"), ("13", "0.9.0")],
    )
    def test_should_return_correct_version_with_v(self):
        assert get_previous_version("0.10.0") == "0.9.0"

    @mock.patch(
        "semantic_release.history.get_commit_log",
        lambda: [("211", "0.10.0-beta"), ("13", "0.9.0")],
    )
    def test_should_return_correct_version_from_prerelease(self):
        assert get_previous_version("0.10.0-beta") == "0.9.0"

    @mock.patch(
        "semantic_release.history.get_commit_log",
        lambda: [("211", "0.10.0"), ("13", "0.10.0-beta"), ("13", "0.9.0")],
    )
    def test_should_return_correct_version_skip_prerelease(self):
        assert get_previous_version("0.10.0-beta", omit_pattern="-beta") == "0.9.0"


class TestGetNewVersion:
    def test_major_bump(self):
        assert get_new_version("0.0.0", "major") == "1.0.0"
        assert get_new_version("0.1.0", "major") == "1.0.0"
        assert get_new_version("0.1.9", "major") == "1.0.0"
        assert get_new_version("10.1.0", "major") == "11.0.0"

    def test_minor_bump(self):
        assert type(get_new_version("0.0.0", "minor")) is str
        assert get_new_version("0.0.0", "minor") == "0.1.0"
        assert get_new_version("1.2.0", "minor") == "1.3.0"
        assert get_new_version("1.2.1", "minor") == "1.3.0"
        assert get_new_version("10.1.0", "minor") == "10.2.0"

    def test_patch_bump(self):
        assert get_new_version("0.0.0", "patch") == "0.0.1"
        assert get_new_version("0.1.0", "patch") == "0.1.1"
        assert get_new_version("10.0.9", "patch") == "10.0.10"

    def test_none_bump(self):
        assert get_new_version("1.0.0", None) == "1.0.0"

    def test_prerelease(self):
        assert get_new_version("1.0.0", None, True) == "1.0.0-beta.0"
        assert get_new_version("1.0.0", "major", True) == "2.0.0-beta.0"
        assert get_new_version("1.0.0", "minor", True) == "1.1.0-beta.0"
        assert get_new_version("1.0.0", "patch", True) == "1.0.1-beta.0"

    def test_prerelease_bump(self, mocker):
        mocker.patch(
            "semantic_release.history.get_current_version", return_value="1.0.0-beta.0"
        )
        assert get_new_version("1.0.0", None, True) == "1.0.0-beta.1"


@mock.patch(
    "semantic_release.history.config.get",
    wrapped_config_get(version_variable="my_version_path:my_version_var"),
)
def test_set_version(tmp_cwd):
    path = tmp_cwd / "my_version_path"
    path.write_text("my_version_var = '1.2.3'")

    set_new_version("X.Y.Z")

    assert path.read_text() == "my_version_var = 'X.Y.Z'"


class TestVersionPattern:
    @pytest.mark.parametrize(
        "str, path, pattern",
        [
            (
                "path:__version__",
                Path("path"),
                r'__version__ *[:=] *["\'](\d+\.\d+(?:\.\d+)?)["\']',
            ),
        ],
    )
    def test_from_variable(self, str, path, pattern):
        p = VersionDeclaration.from_variable(str)
        assert p.path == path
        assert p.pattern == pattern

    @pytest.mark.parametrize(
        "str, path, pattern",
        [
            ("path:pattern", Path("path"), r"pattern"),
            ("path:Version: {version}", Path("path"), r"Version: (\d+\.\d+(?:\.\d+)?)"),
        ],
    )
    def test_from_pattern(self, str, path, pattern):
        p = VersionDeclaration.from_pattern(str)
        assert p.path == path
        assert p.pattern == pattern

    @pytest.mark.parametrize(
        "str, path, key",
        [
            ("path:some.toml.key", Path("path"), r"some.toml.key"),
            ("path:some:other:toml.key", Path("path"), r"some:other:toml.key"),
        ],
    )
    def test_from_toml(self, str, path, key):
        p = VersionDeclaration.from_toml(str)
        assert p.path == path
        assert p.key == key

    @pytest.mark.parametrize(
        "pattern, content, hits",
        [
            (r"(\d+)", "", set()),
            (r"(\d+)", "ab12", {"12"}),
            (r"(\d+)", "ab12 cd34", {"12", "34"}),
            (
                r"version = (\d+)",
                "version = 12\nnotversion = 34\nversion = 56not",
                {"12", "34", "56"},
            ),
            (
                r"^version = (\d+)$",
                "version = 12\nnotversion = 34\nversion = 56not",
                {"12"},
            ),
        ],
    )
    def test_pattern_parse(self, tmp_path, pattern, content, hits):
        path = tmp_path / "pyproject.toml"
        path.write_text(content)

        declaration = PatternVersionDeclaration(str(path), pattern)
        assert declaration.parse() == hits

    @pytest.mark.parametrize(
        "pattern, old_content, new_content",
        [
            (r"(\d+)", "", ""),
            (r"(\d+)", "1", "-"),
            (r"(\d+)", "1b", "-b"),
            (r"(\d+)", "12", "-"),
            (r"(\d+)", "12b", "-b"),
            (r"(\d+)", "a", "a"),
            (r"(\d+)", "a1", "a-"),
            (r"(\d+)", "a1b", "a-b"),
            (r"(\d+)", "a12", "a-"),
            (r"(\d+)", "a12b", "a-b"),
            (r"(\d+)", "a12b3", "a-b-"),
            (r"(\d+)", "a12b3c", "a-b-c"),
            (r"(\d+)", "a12b34", "a-b-"),
            (r"(\d+)", "a12b34c", "a-b-c"),
            (r"a(\d+)", "a", "a"),
            (r"a(\d+)", "a1", "a-"),
            (r"a(\d+)", "a1b", "a-b"),
            (r"a(\d+)", "a12", "a-"),
            (r"a(\d+)", "a12b", "a-b"),
            (r"(\d+)b", "a", "a"),
            (r"(\d+)b", "a1", "a1"),
            (r"(\d+)b", "a1b", "a-b"),
            (r"(\d+)b", "a12", "a12"),
            (r"(\d+)b", "a12b", "a-b"),
        ],
    )
    def test_pattern_replace(self, tmp_path, pattern, old_content, new_content):
        path = tmp_path / "pyproject.toml"
        path.write_text(old_content)

        declaration = PatternVersionDeclaration(str(path), pattern)
        declaration.replace("-")

        assert path.read_text() == new_content

    @pytest.mark.parametrize(
        "key, content, hits",
        [
            ("root", 'root = "test"', {"test"}),
            ("tool.poetry.version", '[tool.poetry]\nversion = "0.1.0"', {"0.1.0"}),
        ],
    )
    def test_toml_parse(self, tmp_path, key, content, hits):
        path = tmp_path / "pyproject.toml"
        path.write_text(content)

        declaration = TomlVersionDeclaration(path, key)
        assert declaration.parse() == hits

    @pytest.mark.parametrize(
        "key, old_content, new_content",
        [
            (r"root", "", ""),
            (r"root", 'root = "test"', 'root = "-"'),
            (
                "tool.poetry.version",
                dedent(
                    """
                    [tool.poetry]
                    version = "0.1.0"
                    [tool.poetry.dependencies.pylint]
                    version = "^2.5.3"
                    optional = true
                    """
                ),
                dedent(
                    """
                    [tool.poetry]
                    version = "-"
                    [tool.poetry.dependencies.pylint]
                    version = "^2.5.3"
                    optional = true
                    """
                ),
            ),
            (
                "tool.poetry.version",
                dedent(
                    """
                    [tool.poetry]
                    name = "my-package"
                    version = "0.1.0"
                    description = "A super package"

                    [build-system]
                    requires = ["poetry-core>=1.0.0"]
                    build-backend = "poetry.core.masonry.api"

                    [tool.semantic_release]
                    version_toml = "pyproject.toml:tool.poetry.version"
                    """
                ),
                dedent(
                    """
                    [tool.poetry]
                    name = "my-package"
                    version = "-"
                    description = "A super package"

                    [build-system]
                    requires = ["poetry-core>=1.0.0"]
                    build-backend = "poetry.core.masonry.api"

                    [tool.semantic_release]
                    version_toml = "pyproject.toml:tool.poetry.version"
                    """
                ),
            ),
        ],
    )
    def test_toml_replace(self, tmp_path, key, old_content, new_content):
        path = tmp_path / "pyproject.toml"
        path.write_text(old_content)

        declaration = TomlVersionDeclaration(str(path), key)
        declaration.replace("-")

        assert path.read_text() == new_content


@pytest.mark.parametrize(
    "params",
    [
        dict(
            pyproject="",
            error=True,
        ),
        dict(
            pyproject="""\
                        [tool.semantic_release]
                        """,
            error=True,
        ),
        dict(
            pyproject="""\
                        [tool.semantic_release]
                        version_variable = "path:__version__"
                        """,
            patterns=[
                (Path("path"), r'__version__ *[:=] *["\'](\d+\.\d+(?:\.\d+)?)["\']'),
            ],
        ),
        dict(
            pyproject="""\
                        [tool.semantic_release]
                        version_variable = "path1:var1,path2:var2"
                        """,
            patterns=[
                (Path("path1"), r'var1 *[:=] *["\'](\d+\.\d+(?:\.\d+)?)["\']'),
                (Path("path2"), r'var2 *[:=] *["\'](\d+\.\d+(?:\.\d+)?)["\']'),
            ],
        ),
        dict(
            pyproject="""\
                        [tool.semantic_release]
                        version_variable = [
                            "path1:var1",
                            "path2:var2"
                        ]
                        """,
            patterns=[
                (Path("path1"), r'var1 *[:=] *["\'](\d+\.\d+(?:\.\d+)?)["\']'),
                (Path("path2"), r'var2 *[:=] *["\'](\d+\.\d+(?:\.\d+)?)["\']'),
            ],
        ),
        dict(
            pyproject="""\
                        [tool.semantic_release]
                        version_pattern = "path:pattern"
                        """,
            patterns=[
                (Path("path"), "pattern"),
            ],
        ),
        dict(
            pyproject="""\
                        [tool.semantic_release]
                        version_pattern = "path1:pattern1,path2:pattern2"
                        """,
            patterns=[
                (Path("path1"), "pattern1"),
                (Path("path2"), "pattern2"),
            ],
        ),
        dict(
            pyproject="""\
                        [tool.semantic_release]
                        version_pattern = [
                            "path1:pattern1",
                            "path2:pattern2"
                        ]
                        """,
            patterns=[
                (Path("path1"), "pattern1"),
                (Path("path2"), "pattern2"),
            ],
        ),
    ],
)
def test_load_version_patterns(tmp_cwd, monkeypatch, params):
    import semantic_release.history
    import semantic_release.settings

    config = tmp_cwd / "pyproject.toml"
    config.write_text(dedent(params["pyproject"]))
    print(config.read_text())

    monkeypatch.setattr(
        semantic_release.history,
        "config",
        semantic_release.settings._config(),
    )

    if "error" in params:
        with pytest.raises(ImproperConfigurationError):
            load_version_declarations()

    else:
        patterns = load_version_declarations()
        pattern_tuples = [(x.path, x.pattern) for x in patterns]
        assert pattern_tuples == params["patterns"]



"""CI Checks
"""
import os
from typing import Callable

from semantic_release.errors import CiVerificationError


def checker(func: Callable) -> Callable:
    """
    A decorator that will convert AssertionErrors into
    CiVerificationError.

    :param func: A function that will raise AssertionError
    :return: The given function wrapped to raise a CiVerificationError on AssertionError
    """

    def func_wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
            return True
        except AssertionError:
            raise CiVerificationError(
                "The verification check for the environment did not pass."
            )

    return func_wrapper


@checker
def travis(branch: str):
    """
    Performs necessary checks to ensure that the travis build is one
    that should create releases.

    :param branch: The branch the environment should be running against.
    """
    assert os.environ.get("TRAVIS_BRANCH") == branch
    assert os.environ.get("TRAVIS_PULL_REQUEST") == "false"


@checker
def semaphore(branch: str):
    """
    Performs necessary checks to ensure that the semaphore build is successful,
    on the correct branch and not a pull-request.

    :param branch:  The branch the environment should be running against.
    """
    assert os.environ.get("BRANCH_NAME") == branch
    assert os.environ.get("PULL_REQUEST_NUMBER") is None
    assert os.environ.get("SEMAPHORE_THREAD_RESULT") != "failed"


@checker
def frigg(branch: str):
    """
    Performs necessary checks to ensure that the frigg build is one
    that should create releases.

    :param branch: The branch the environment should be running against.
    """
    assert os.environ.get("FRIGG_BUILD_BRANCH") == branch
    assert not os.environ.get("FRIGG_PULL_REQUEST")


@checker
def circle(branch: str):
    """
    Performs necessary checks to ensure that the circle build is one
    that should create releases.

    :param branch: The branch the environment should be running against.
    """
    assert os.environ.get("CIRCLE_BRANCH") == branch
    assert not os.environ.get("CI_PULL_REQUEST")


@checker
def gitlab(branch: str):
    """
    Performs necessary checks to ensure that the gitlab build is one
    that should create releases.

    :param branch: The branch the environment should be running against.
    """
    assert os.environ.get("CI_COMMIT_REF_NAME") == branch
    # TODO - don't think there's a merge request indicator variable


@checker
def bitbucket(branch: str):
    """
    Performs necessary checks to ensure that the bitbucket build is one
    that should create releases.

    :param branch: The branch the environment should be running against.
    """
    assert os.environ.get("BITBUCKET_BRANCH") == branch
    assert not os.environ.get("BITBUCKET_PR_ID")


@checker
def jenkins(branch: str):
    """
    Performs necessary checks to ensure that the jenkins build is one
    that should create releases.

    :param branch: The branch the environment should be running against.
    """

    branch_name = os.environ.get("BRANCH_NAME") or os.environ.get("GIT_BRANCH")
    assert os.environ.get("JENKINS_URL") is not None
    assert branch_name == branch
    assert not os.environ.get("CHANGE_ID")  # pull request id


def check(branch: str = "master"):
    """
    Detects the current CI environment, if any, and performs necessary
    environment checks.

    :param branch: The branch that should be the current branch.
    """
    if os.environ.get("TRAVIS") == "true":
        travis(branch)
    elif os.environ.get("SEMAPHORE") == "true":
        semaphore(branch)
    elif os.environ.get("FRIGG") == "true":
        frigg(branch)
    elif os.environ.get("CIRCLECI") == "true":
        circle(branch)
    elif os.environ.get("GITLAB_CI") == "true":
        gitlab(branch)
    elif os.environ.get("JENKINS_URL") is not None:
        jenkins(branch)
    elif "BITBUCKET_BUILD_NUMBER" in os.environ:
        bitbucket(branch)



import pytest

from semantic_release import ci_checks
from semantic_release.errors import CiVerificationError


def test_travis_should_pass_if_branch_is_master_and_no_pr(monkeypatch):
    monkeypatch.setenv("TRAVIS_BRANCH", "master")
    monkeypatch.setenv("TRAVIS_PULL_REQUEST", "false")

    assert ci_checks.travis("master")


def test_travis_should_pass_if_branch_is_correct_and_no_pr(monkeypatch):
    monkeypatch.setenv("TRAVIS_BRANCH", "other-branch")
    monkeypatch.setenv("TRAVIS_PULL_REQUEST", "false")

    assert ci_checks.travis("other-branch")


def test_travis_should_raise_ci_verification_error_for_wrong_branch(monkeypatch):
    monkeypatch.setenv("TRAVIS_BRANCH", "other-branch")
    monkeypatch.setenv("TRAVIS_PULL_REQUEST", "false")

    with pytest.raises(CiVerificationError):
        ci_checks.travis("master")


def test_travis_should_raise_ci_verification_error_for_pr(monkeypatch):
    monkeypatch.setenv("TRAVIS_BRANCH", "other-branch")
    monkeypatch.setenv("TRAVIS_PULL_REQUEST", "42")

    with pytest.raises(CiVerificationError):
        ci_checks.travis("master")



from semantic_release import ci_checks


def test_check_should_call_travis_with_correct_env_variable(mocker, monkeypatch):
    mock_travis = mocker.patch("semantic_release.ci_checks.travis")
    monkeypatch.setenv("TRAVIS", "true")
    ci_checks.check("master")
    mock_travis.assert_called_once_with("master")


def test_check_should_call_semaphore_with_correct_env_variable(mocker, monkeypatch):
    mock_semaphore = mocker.patch("semantic_release.ci_checks.semaphore")
    monkeypatch.setenv("SEMAPHORE", "true")
    ci_checks.check("master")
    mock_semaphore.assert_called_once_with("master")


def test_check_should_call_frigg_with_correct_env_variable(mocker, monkeypatch):
    mock_frigg = mocker.patch("semantic_release.ci_checks.frigg")
    monkeypatch.setenv("FRIGG", "true")
    ci_checks.check("master")
    mock_frigg.assert_called_once_with("master")


def test_check_should_call_circle_with_correct_env_variable(mocker, monkeypatch):
    mock_circle = mocker.patch("semantic_release.ci_checks.circle")
    monkeypatch.setenv("CIRCLECI", "true")
    ci_checks.check("master")
    mock_circle.assert_called_once_with("master")


def test_check_should_call_gitlab_with_correct_env_variable(mocker, monkeypatch):
    mock_gitlab = mocker.patch("semantic_release.ci_checks.gitlab")
    monkeypatch.setenv("GITLAB_CI", "true")
    ci_checks.check("master")
    mock_gitlab.assert_called_once_with("master")


def test_check_should_call_bitbucket_with_correct_env_variable(mocker, monkeypatch):
    mock_bitbucket = mocker.patch("semantic_release.ci_checks.bitbucket")
    monkeypatch.setenv("BITBUCKET_BUILD_NUMBER", "true")
    ci_checks.check("master")
    mock_bitbucket.assert_called_once_with("master")



import pytest

from semantic_release import ci_checks
from semantic_release.errors import CiVerificationError


def test_semaphore_should_pass_if_branch_and_status_is_ok(monkeypatch):
    monkeypatch.setenv("BRANCH_NAME", "master")
    monkeypatch.setenv("SEMAPHORE_THREAD_RESULT", "success")

    assert ci_checks.semaphore("master")


def test_semaphore_should_raise_ci_verification_error_if_branch_is_wrong(monkeypatch):
    monkeypatch.setenv("BRANCH_NAME", "other-branch")
    monkeypatch.setenv("SEMAPHORE_THREAD_RESULT", "success")

    with pytest.raises(CiVerificationError):
        ci_checks.semaphore("master")


def test_semaphore_should_raise_ci_verification_error_if_status_is_failed(monkeypatch):
    monkeypatch.setenv("BRANCH_NAME", "master")
    monkeypatch.setenv("SEMAPHORE_THREAD_RESULT", "failed")

    with pytest.raises(CiVerificationError):
        ci_checks.semaphore("master")


def test_semaphore_should_raise_ci_verification_error_if_pull_request_number_is_none(
    monkeypatch,
):
    monkeypatch.setenv("BRANCH_NAME", "master")
    monkeypatch.setenv("SEMAPHORE_THREAD_RESULT", "success")
    monkeypatch.setenv("PULL_REQUEST_NUMBER", "42")

    with pytest.raises(CiVerificationError):
        ci_checks.semaphore("master")



from semantic_release.history import emoji_parser


def test_major():
    commit = (
        ":boom: Breaking changes\n\n" "More description\n\n" "Even more description"
    )
    parsed_commit = emoji_parser(commit)
    assert parsed_commit[0] == 3
    assert parsed_commit[1] == ":boom:"
    assert parsed_commit[3] == [
        ":boom: Breaking changes",
        "More description",
        "Even more description",
    ]
    assert parsed_commit[4] == ["More description", "Even more description"]


def test_minor():
    commit = ":sparkles: Add a new feature\n\n" "Some description of the feature"
    parsed_commit = emoji_parser(commit)
    assert parsed_commit[0] == 2
    assert parsed_commit[1] == ":sparkles:"
    assert parsed_commit[3] == [
        ":sparkles: Add a new feature",
        "Some description of the feature",
    ]
    assert parsed_commit[4] == []


def test_patch():
    commit = ":bug: Fixing a bug\n\n" "The bug is finally gone!"
    parsed_commit = emoji_parser(commit)
    assert parsed_commit[0] == 1
    assert parsed_commit[1] == ":bug:"
    assert parsed_commit[3] == [":bug: Fixing a bug", "The bug is finally gone!"]
    assert parsed_commit[4] == []


def test_other_emoji():
    commit = ":pencil: Documentation changes"
    parsed_commit = emoji_parser(commit)
    assert parsed_commit[0] == 0
    assert parsed_commit[1] == "Other"
    assert parsed_commit[3] == [":pencil: Documentation changes"]
    assert parsed_commit[4] == []


def test_multiple_emojis():
    commit = ":sparkles::pencil: Add a feature and document it"
    parsed_commit = emoji_parser(commit)
    assert parsed_commit[0] == 2
    assert parsed_commit[1] == ":sparkles:"
    assert parsed_commit[3] == [":sparkles::pencil: Add a feature and document it"]
    assert parsed_commit[4] == []


def test_emoji_in_description():
    commit = ":sparkles: Add a new feature\n\n" ":boom: should not be detected"
    parsed_commit = emoji_parser(commit)
    assert parsed_commit[0] == 2
    assert parsed_commit[1] == ":sparkles:"
    assert parsed_commit[3] == [
        ":sparkles: Add a new feature",
        ":boom: should not be detected",
    ]
    assert parsed_commit[4] == []



import mock

from semantic_release.history import evaluate_version_bump

from .. import wrapped_config_get
from . import *


def test_major():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: ALL_KINDS_OF_COMMIT_MESSAGES,
    ):
        assert evaluate_version_bump("0.0.0") == "major"


def test_minor():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: MINOR_AND_PATCH_COMMIT_MESSAGES,
    ):
        assert evaluate_version_bump("0.0.0") == "minor"


def test_patch():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: PATCH_COMMIT_MESSAGES,
    ):
        assert evaluate_version_bump("0.0.0") == "patch"


def test_nothing_if_no_tag():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: [("", "...")],
    ):
        assert evaluate_version_bump("0.0.0") is None


def test_force():
    assert evaluate_version_bump("0.0.0", "major") == "major"
    assert evaluate_version_bump("0.0.0", "minor") == "minor"
    assert evaluate_version_bump("0.0.0", "patch") == "patch"


def test_should_not_skip_commits_mentioning_other_commits():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: MAJOR_MENTIONING_LAST_VERSION,
    ):
        assert evaluate_version_bump("1.0.0") == "major"


@mock.patch(
    "semantic_release.history.config.get", wrapped_config_get(patch_without_tag=True)
)
@mock.patch("semantic_release.history.logs.get_commit_log", lambda *a, **kw: [MINOR])
def test_should_minor_with_patch_without_tag():
    assert evaluate_version_bump("1.1.0") == "minor"


@mock.patch(
    "semantic_release.history.config.get", wrapped_config_get(patch_without_tag=True)
)
@mock.patch("semantic_release.history.logs.get_commit_log", lambda *a, **kw: [NO_TAG])
def test_should_patch_without_tagged_commits():
    assert evaluate_version_bump("1.1.0") == "patch"


@mock.patch(
    "semantic_release.history.config.get", wrapped_config_get(patch_without_tag=False)
)
@mock.patch("semantic_release.history.logs.get_commit_log", lambda *a, **kw: [NO_TAG])
def test_should_return_none_without_tagged_commits():
    assert evaluate_version_bump("1.1.0") is None


@mock.patch("semantic_release.history.logs.get_commit_log", lambda *a, **kw: [])
def test_should_return_none_without_commits():
    """
    Make sure that we do not release if there are no commits since last release.
    """
    with mock.patch("semantic_release.history.config.get", lambda *x: True):
        assert evaluate_version_bump("1.1.0") is None

    with mock.patch("semantic_release.history.config.get", lambda *x: False):
        assert evaluate_version_bump("1.1.0") is None


@mock.patch(
    "semantic_release.history.config.get", wrapped_config_get(major_on_zero=False)
)
@mock.patch("semantic_release.history.logs.get_commit_log", lambda *a, **kw: [MAJOR])
def test_should_minor_without_major_on_zero():
    assert evaluate_version_bump("0.1.0") == "minor"



import pytest

from semantic_release.errors import UnknownCommitMessageStyleError
from semantic_release.history import tag_parser

text = (
    "This is an long explanatory part of a commit message. It should give "
    "some insight to the fix this commit adds to the codebase."
)
footer = "Closes #400"


def test_parser_raises_unknown_message_style():
    pytest.raises(UnknownCommitMessageStyleError, tag_parser, "")


def test_parser_return_major_bump_level():
    commit = ":sparkles: Add new parser pattern\n\nBREAKING CHANGE:"
    assert tag_parser(commit)[0] == 3


def test_parser_return_minor_bump_level():
    assert tag_parser(":sparkles: Add emoji parser")[0] == 2


def test_parser_return_patch_bump_level():
    assert tag_parser(":nut_and_bolt: Fix regex in angular parser")[0] == 1


def test_parser_return_type_breaking_from_commit_message():
    commit = ":sparkles: Add new parser pattern\n\nBREAKING CHANGE:"
    assert tag_parser(commit)[1] == "breaking"


def test_parser_return_type_feature_from_commit_message():
    assert tag_parser(":sparkles: ...")[1] == "feature"


def test_parser_return_type_fix_from_commit_message():
    assert tag_parser(":nut_and_bolt: ...")[1] == "fix"


def test_parser_return_subject_from_commit_message():
    assert tag_parser(":sparkles: Add emoji parser")[3][0] == "Add emoji parser"


def test_parser_return_text_from_commit_message():
    commit = ":nut_and_bolt: Fix regex in an parser\n\n{}".format(text)
    assert tag_parser(commit)[3][1] == text


def test_parser_return_footer_from_commit_message():
    commit = ":nut_and_bolt: Fix env \n\n{t[text]}\n\n{t[footer]}".format(t=globals())
    assert tag_parser(commit)[3][2] == footer



import mock
import pytest

from semantic_release.history.logs import generate_changelog

from .. import wrapped_config_get
from . import *


def test_should_generate_necessary_sections():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **k: ALL_KINDS_OF_COMMIT_MESSAGES + [MAJOR2, UNKNOWN_STYLE],
    ):
        changelog = generate_changelog("0.0.0")
        assert len(changelog["feature"]) > 0
        assert len(changelog["fix"]) > 0
        assert len(changelog["breaking"]) > 0
        assert "documentation" not in changelog


def test_should_include_hash_in_section_contents():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **k: ALL_KINDS_OF_COMMIT_MESSAGES,
    ):
        changelog = generate_changelog("0.0.0")
        assert changelog["breaking"][0][0] == MAJOR[0]
        assert changelog["feature"][0][0] == MINOR[0]
        assert changelog["fix"][0][0] == PATCH[0]


def test_should_only_read_until_given_version():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **k: MAJOR_LAST_RELEASE_MINOR_AFTER,
    ):
        changelog = generate_changelog("1.1.0")
        assert len(changelog["breaking"]) == 0
        assert len(changelog["feature"]) > 0
        assert "fix" not in changelog
        assert "documentation" not in changelog


@pytest.mark.parametrize(
    "commit,expected_description",
    [
        (MAJOR, "Uses super-feature as default instead of dull-feature."),
        (MAJOR2, "Uses super-feature as default instead of dull-feature."),
        (
            MAJOR_MENTIONING_1_0_0,
            "Uses super-feature as default instead of dull-feature from v1.0.0.",
        ),
        (MAJOR_EXCL_WITH_FOOTER, "Another feature, another breaking change"),
        (MAJOR_EXCL_NOT_FOOTER, "Fix a big bug that everyone exploited"),
    ],
)
def test_should_get_right_breaking_description(commit, expected_description):
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: [commit],
    ):
        changelog = generate_changelog("0.0.0")
        assert changelog["breaking"][0][1] == expected_description


def test_should_get_multiple_breaking_descriptions():
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: [MAJOR_MULTIPLE_FOOTERS],
    ):
        changelog = generate_changelog("0.0.0")
        assert len(changelog["breaking"]) == 2
        assert changelog["breaking"][0][1] == "Breaking change 1"
        assert changelog["breaking"][1][1] == "Breaking change 2"


@pytest.mark.parametrize(
    "commit,commit_type,expected",
    [
        (MAJOR, "feature", "**x:** Add super-feature"),
        (MINOR, "feature", "**x:** Add non-breaking super-feature"),
        (PATCH, "fix", "**x:** Fix bug in super-feature"),
    ],
)
def test_scope_is_included_in_changelog(commit, commit_type, expected):
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: [commit],
    ):
        changelog = generate_changelog("0.0.0")
        assert changelog[commit_type][0][1] == expected


@mock.patch(
    "semantic_release.history.logs.get_commit_log",
    lambda *a, **k: [("24", "fix: Fix another bug")],
)
def test_scope_is_omitted_with_empty_scope():
    changelog = generate_changelog("0.0.0")
    assert changelog["fix"][0][1] == "Fix another bug"


@mock.patch(
    "semantic_release.history.config.get", wrapped_config_get(changelog_scope=False)
)
@pytest.mark.parametrize(
    "commit,commit_type",
    [
        (MAJOR, "feature"),
        (MINOR, "feature"),
        (PATCH, "fix"),
    ],
)
def test_scope_included_in_changelog_configurable(commit, commit_type):
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: [commit],
    ):
        changelog = generate_changelog("0.0.0")
        assert "**x:**" not in changelog[commit_type][0][1]


@mock.patch(
    "semantic_release.history.logs.get_commit_log",
    lambda *a, **k: [("23", "fix(x): abCD"), ("23", "fix: abCD")],
)
@pytest.mark.parametrize(
    "commit,config_setting,expected_description",
    [
        (("23", "fix(x): abCD"), True, "**x:** AbCD"),
        (("23", "fix(x): abCD"), False, "**x:** abCD"),
        (("23", "fix: abCD"), True, "AbCD"),
        (("23", "fix: abCD"), False, "abCD"),
    ],
)
def test_message_capitalization_is_configurable(
    commit, config_setting, expected_description
):
    with mock.patch(
        "semantic_release.history.logs.get_commit_log",
        lambda *a, **kw: [commit],
    ):

        with mock.patch(
            "semantic_release.history.config.get",
            wrapped_config_get(changelog_capitalize=config_setting),
        ):
            changelog = generate_changelog("0.0.0")
            assert changelog["fix"][0][1] == expected_description



"""HVCS
"""
import logging
import mimetypes
import os
from typing import Any, Optional, Union, cast
from urllib.parse import urlsplit

from gitlab import exceptions, gitlab
from requests import HTTPError, Session
from requests.auth import AuthBase
from urllib3 import Retry

from .errors import ImproperConfigurationError
from .helpers import LoggedFunction, build_requests_session
from .settings import config
from .vcs_helpers import get_formatted_tag

logger = logging.getLogger(__name__)

# Add a mime type for wheels
mimetypes.add_type("application/octet-stream", ".whl")


class Base(object):
    @staticmethod
    def domain() -> str:
        raise NotImplementedError

    @staticmethod
    def api_url() -> str:
        raise NotImplementedError

    @staticmethod
    def token() -> Optional[str]:
        raise NotImplementedError

    @staticmethod
    def check_build_status(owner: str, repo: str, ref: str) -> bool:
        raise NotImplementedError

    @classmethod
    def post_release_changelog(
        cls, owner: str, repo: str, version: str, changelog: str
    ) -> bool:
        raise NotImplementedError

    @classmethod
    def upload_dists(cls, owner: str, repo: str, version: str, path: str) -> bool:
        # Skip on unsupported HVCS instead of raising error
        return True


def _fix_mime_types():
    """Fix incorrect entries in the `mimetypes` registry.
    On Windows, the Python standard library's `mimetypes` reads in
    mappings from file extension to MIME type from the Windows
    registry. Other applications can and do write incorrect values
    to this registry, which causes `mimetypes.guess_type` to return
    incorrect values, which causes TensorBoard to fail to render on
    the frontend.
    This method hard-codes the correct mappings for certain MIME
    types that are known to be either used by python-semantic-release or
    problematic in general.
    """
    mimetypes.add_type("text/markdown", ".md")


class TokenAuth(AuthBase):
    """
    requests Authentication for token based authorization
    """

    def __init__(self, token):
        self.token = token

    def __eq__(self, other):
        return all(
            [
                self.token == getattr(other, "token", None),
            ]
        )

    def __ne__(self, other):
        return not self == other

    def __call__(self, r):
        r.headers["Authorization"] = f"token {self.token}"
        return r


class Github(Base):
    """Github helper class"""

    DEFAULT_DOMAIN = "github.com"
    _fix_mime_types()

    @staticmethod
    def domain() -> str:
        """Github domain property

        :return: The Github domain
        """
        # ref: https://docs.github.com/en/actions/reference/environment-variables#default-environment-variables
        hvcs_domain = config.get(
            "hvcs_domain", os.getenv("GITHUB_SERVER_URL", "").replace("https://", "")
        )
        domain = hvcs_domain if hvcs_domain else Github.DEFAULT_DOMAIN
        return domain

    @staticmethod
    def api_url() -> str:
        """Github api_url property

        :return: The Github API URL
        """
        # not necessarily prefixed with api in the case of a custom domain, so
        # can't just default DEFAULT_DOMAIN to github.com
        # ref: https://docs.github.com/en/actions/reference/environment-variables#default-environment-variables
        hvcs_api_domain = config.get(
            "hvcs_api_domain", os.getenv("GITHUB_API_URL", "").replace("https://", "")
        )
        hostname = (
            hvcs_api_domain if hvcs_api_domain else "api." + Github.DEFAULT_DOMAIN
        )
        return f"https://{hostname}"

    @staticmethod
    def token() -> Optional[str]:
        """Github token property

        :return: The Github token environment variable (GH_TOKEN) value
        """
        return os.environ.get(config.get("github_token_var"))

    @staticmethod
    def auth() -> Optional[TokenAuth]:
        """Github token property

        :return: The Github token environment variable (GH_TOKEN) value
        """
        token = Github.token()
        if not token:
            return None
        return TokenAuth(token)

    @staticmethod
    def session(
        raise_for_status=True, retry: Union[Retry, bool, int] = True
    ) -> Session:
        session = build_requests_session(raise_for_status=raise_for_status, retry=retry)
        session.auth = Github.auth()
        return session

    @staticmethod
    @LoggedFunction(logger)
    def check_build_status(owner: str, repo: str, ref: str) -> bool:
        """Check build status

        https://docs.github.com/rest/reference/repos#get-the-combined-status-for-a-specific-reference

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param ref: The sha1 hash of the commit ref

        :return: Was the build status success?
        """
        url = "{domain}/repos/{owner}/{repo}/commits/{ref}/status"
        try:
            response = Github.session().get(
                url.format(domain=Github.api_url(), owner=owner, repo=repo, ref=ref)
            )
            return response.json().get("state") == "success"
        except HTTPError as e:
            logger.warning(f"Build status check on Github has failed: {e}")
            return False

    @classmethod
    @LoggedFunction(logger)
    def create_release(cls, owner: str, repo: str, tag: str, changelog: str) -> bool:
        """Create a new release

        https://docs.github.com/rest/reference/repos#create-a-release

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param tag: Tag to create release for
        :param changelog: The release notes for this version

        :return: Whether the request succeeded
        """
        try:
            Github.session().post(
                f"{Github.api_url()}/repos/{owner}/{repo}/releases",
                json={
                    "tag_name": tag,
                    "name": tag,
                    "body": changelog,
                    "draft": False,
                    "prerelease": False,
                },
            )
            return True
        except HTTPError as e:
            logger.warning(f"Release creation on Github has failed: {e}")
            return False

    @classmethod
    @LoggedFunction(logger)
    def get_release(cls, owner: str, repo: str, tag: str) -> Optional[int]:
        """Get a release by its tag name

        https://docs.github.com/rest/reference/repos#get-a-release-by-tag-name

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param tag: Tag to get release for

        :return: ID of found release
        """
        try:
            response = Github.session().get(
                f"{Github.api_url()}/repos/{owner}/{repo}/releases/tags/{tag}"
            )
            return response.json().get("id")
        except HTTPError as e:
            if e.response.status_code != 404:
                logger.debug(f"Get release by tag on Github has failed: {e}")
            return None

    @classmethod
    @LoggedFunction(logger)
    def edit_release(cls, owner: str, repo: str, id: int, changelog: str) -> bool:
        """Edit a release with updated change notes

        https://docs.github.com/rest/reference/repos#update-a-release

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param id: ID of release to update
        :param changelog: The release notes for this version

        :return: Whether the request succeeded
        """
        try:
            Github.session().post(
                f"{Github.api_url()}/repos/{owner}/{repo}/releases/{id}",
                json={"body": changelog},
            )
            return True
        except HTTPError as e:
            logger.warning(f"Edit release on Github has failed: {e}")
            return False

    @classmethod
    @LoggedFunction(logger)
    def post_release_changelog(
        cls, owner: str, repo: str, version: str, changelog: str
    ) -> bool:
        """Post release changelog

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param version: The version number
        :param changelog: The release notes for this version

        :return: The status of the request
        """
        tag = get_formatted_tag(version)
        logger.debug(f"Attempting to create release for {tag}")
        success = Github.create_release(owner, repo, tag, changelog)

        if not success:
            logger.debug("Unsuccessful, looking for an existing release to update")
            release_id = Github.get_release(owner, repo, tag)
            if release_id:
                logger.debug(f"Updating release {release_id}")
                success = Github.edit_release(owner, repo, release_id, changelog)
            else:
                logger.debug(f"Existing release not found")

        return success

    @classmethod
    @LoggedFunction(logger)
    def upload_asset(
        cls, owner: str, repo: str, release_id: int, file: str, label: str = None
    ) -> bool:
        """Upload an asset to an existing release

        https://docs.github.com/rest/reference/repos#upload-a-release-asset

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param release_id: ID of the release to upload to
        :param file: Path of the file to upload
        :param label: Custom label for this file

        :return: The status of the request
        """
        url = f"https://uploads.github.com/repos/{owner}/{repo}/releases/{release_id}/assets"

        content_type = mimetypes.guess_type(file, strict=False)[0]
        if not content_type:
            content_type = "application/octet-stream"

        try:
            response = Github.session().post(
                url,
                params={"name": os.path.basename(file), "label": label},
                headers={
                    "Content-Type": content_type,
                },
                data=open(file, "rb").read(),
            )

            logger.debug(
                f"Asset upload on Github completed, url: {response.url}, status code: {response.status_code}"
            )

            return True
        except HTTPError as e:
            logger.warning(f"Asset upload {file} on Github has failed: {e}")
            return False

    @classmethod
    def upload_dists(cls, owner: str, repo: str, version: str, path: str) -> bool:
        """Upload distributions to a release

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param version: Version to upload for
        :param path: Path to the dist directory

        :return: The status of the request
        """

        # Find the release corresponding to this version
        release_id = Github.get_release(owner, repo, get_formatted_tag(version))
        if not release_id:
            logger.debug("No release found to upload assets to")
            return False

        # Upload assets
        one_or_more_failed = False
        for file in os.listdir(path):
            file_path = os.path.join(path, file)

            if not Github.upload_asset(owner, repo, release_id, file_path):
                one_or_more_failed = True

        return not one_or_more_failed


class Gitea(Base):
    """Gitea helper class"""

    DEFAULT_DOMAIN = "gitea.com"
    DEFAULT_API_PATH = "/api/v1"

    _fix_mime_types()

    @staticmethod
    def domain() -> str:
        """Gitea domain property

        :return: The Gitea domain
        """
        # ref: https://docs.gitea.com/en/actions/reference/environment-variables#default-environment-variables
        hvcs_domain = config.get(
            "hvcs_domain", os.getenv("GITEA_SERVER_URL", "").replace("https://", "")
        )
        domain = hvcs_domain if hvcs_domain else Gitea.DEFAULT_DOMAIN
        return domain

    @staticmethod
    def api_url() -> str:
        """Gitea api_url property

        :return: The Gitea API URL
        """

        hvcs_domain = config.get(
            "hvcs_domain", os.getenv("GITEA_SERVER_URL", "").replace("https://", "")
        )
        hostname = config.get(
            "hvcs_api_domain", os.getenv("GITEA_API_URL", "").replace("https://", "")
        )

        if hvcs_domain and not hostname:
            hostname = hvcs_domain + Gitea.DEFAULT_API_PATH
        elif not hostname:
            hostname = Gitea.DEFAULT_DOMAIN + Gitea.DEFAULT_API_PATH

        return f"https://{hostname}"

    @staticmethod
    def token() -> Optional[str]:
        """Gitea token property

        :return: The Gitea token environment variable (GITEA_TOKEN) value
        """
        return os.environ.get(config.get("gitea_token_var"))

    @staticmethod
    def auth() -> Optional[TokenAuth]:
        """Gitea token property

        :return: The Gitea token environment variable (GITEA_TOKEN) value
        """
        token = Gitea.token()
        if not token:
            return None
        return TokenAuth(token)

    @staticmethod
    def session(
        raise_for_status=True, retry: Union[Retry, bool, int] = True
    ) -> Session:
        session = build_requests_session(raise_for_status=raise_for_status, retry=retry)
        session.auth = Gitea.auth()
        return session

    @staticmethod
    @LoggedFunction(logger)
    def check_build_status(owner: str, repo: str, ref: str) -> bool:
        """Check build status

        https://gitea.com/api/swagger#/repository/repoCreateStatus

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param ref: The sha1 hash of the commit ref

        :return: Was the build status success?
        """
        url = "{domain}/repos/{owner}/{repo}/statuses/{ref}"
        try:
            response = Gitea.session().get(
                url.format(domain=Gitea.api_url(), owner=owner, repo=repo, ref=ref)
            )
            data = response.json()
            if type(data) == list:
                return data[0].get("status") == "success"
            else:
                return data.get("status") == "success"
        except HTTPError as e:
            logger.warning(f"Build status check on Gitea has failed: {e}")
            return False

    @classmethod
    @LoggedFunction(logger)
    def create_release(cls, owner: str, repo: str, tag: str, changelog: str) -> bool:
        """Create a new release

        https://gitea.com/api/swagger#/repository/repoCreateRelease

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param tag: Tag to create release for
        :param changelog: The release notes for this version

        :return: Whether the request succeeded
        """
        try:
            Gitea.session().post(
                f"{Gitea.api_url()}/repos/{owner}/{repo}/releases",
                json={
                    "tag_name": tag,
                    "name": tag,
                    "body": changelog,
                    "draft": False,
                    "prerelease": False,
                },
            )
            return True
        except HTTPError as e:
            logger.warning(f"Release creation on Gitea has failed: {e}")
            return False

    @classmethod
    @LoggedFunction(logger)
    def get_release(cls, owner: str, repo: str, tag: str) -> Optional[int]:
        """Get a release by its tag name

        https://gitea.com/api/swagger#/repository/repoGetReleaseByTag

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param tag: Tag to get release for

        :return: ID of found release
        """
        try:
            response = Gitea.session().get(
                f"{Gitea.api_url()}/repos/{owner}/{repo}/releases/tags/{tag}"
            )
            return response.json().get("id")
        except HTTPError as e:
            if e.response.status_code != 404:
                logger.debug(f"Get release by tag on Gitea has failed: {e}")
            return None

    @classmethod
    @LoggedFunction(logger)
    def edit_release(cls, owner: str, repo: str, id: int, changelog: str) -> bool:
        """Edit a release with updated change notes

        https://gitea.com/api/swagger#/repository/repoEditRelease

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param id: ID of release to update
        :param changelog: The release notes for this version

        :return: Whether the request succeeded
        """
        try:
            Gitea.session().post(
                f"{Gitea.api_url()}/repos/{owner}/{repo}/releases/{id}",
                json={"body": changelog},
            )
            return True
        except HTTPError as e:
            logger.warning(f"Edit release on Gitea has failed: {e}")
            return False

    @classmethod
    @LoggedFunction(logger)
    def post_release_changelog(
        cls, owner: str, repo: str, version: str, changelog: str
    ) -> bool:
        """Post release changelog

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param version: The version number
        :param changelog: The release notes for this version

        :return: The status of the request
        """
        tag = get_formatted_tag(version)
        logger.debug(f"Attempting to create release for {tag}")
        success = Gitea.create_release(owner, repo, tag, changelog)

        if not success:
            logger.debug("Unsuccessful, looking for an existing release to update")
            release_id = Gitea.get_release(owner, repo, tag)
            if release_id:
                logger.debug(f"Updating release {release_id}")
                success = Gitea.edit_release(owner, repo, release_id, changelog)
            else:
                logger.debug(f"Existing release not found")

        return success

    @classmethod
    @LoggedFunction(logger)
    def upload_asset(
        cls, owner: str, repo: str, release_id: int, file: str, label: str = None
    ) -> bool:
        """Upload an asset to an existing release

        https://gitea.com/api/swagger#/repository/repoCreateReleaseAttachment

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param release_id: ID of the release to upload to
        :param file: Path of the file to upload
        :param label: Custom label for this file

        :return: The status of the request
        """
        url = f"{Gitea.api_url()}/repos/{owner}/{repo}/releases/{release_id}/assets"
        try:
            name = os.path.basename(file)
            response = Gitea.session().post(
                url,
                params={"name": name},
                data={},
                files=[
                    (
                        "attachment",
                        (
                            name,
                            open(file, "rb"),
                            "application/octet-stream",
                        ),
                    )
                ],
            )

            logger.debug(
                f"Asset upload on Gitea completed, url: {response.url}, status code: {response.status_code}"
            )

            return True
        except HTTPError as e:
            logger.warning(f"Asset upload {file} on Gitea has failed: {e}")
            return False

    @classmethod
    def upload_dists(cls, owner: str, repo: str, version: str, path: str) -> bool:
        """Upload distributions to a release

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param version: Version to upload for
        :param path: Path to the dist directory

        :return: The status of the request
        """

        # Find the release corresponding to this version
        release_id = Gitea.get_release(owner, repo, get_formatted_tag(version))
        if not release_id:
            logger.debug("No release found to upload assets to")
            return False

        # Upload assets
        one_or_more_failed = False
        for file in os.listdir(path):
            file_path = os.path.join(path, file)

            if not Gitea.upload_asset(owner, repo, release_id, file_path):
                one_or_more_failed = True

        return not one_or_more_failed


class Gitlab(Base):
    """Gitlab helper class"""

    @staticmethod
    def domain() -> str:
        """Gitlab domain property

        :return: The Gitlab instance domain
        """
        # Use Gitlab-CI environment vars if available
        if "CI_SERVER_URL" in os.environ:
            url = urlsplit(os.environ["CI_SERVER_URL"])
            return f"{url.netloc}{url.path}".rstrip("/")

        domain = config.get("hvcs_domain", os.environ.get("CI_SERVER_HOST", None))
        return domain if domain else "gitlab.com"

    @staticmethod
    def api_url() -> str:
        """Gitlab api_url property

        :return: The Gitlab instance API url
        """
        # Use Gitlab-CI environment vars if available
        if "CI_SERVER_URL" in os.environ:
            return os.environ["CI_SERVER_URL"]

        return f"http://{Gitlab.domain()}"

    @staticmethod
    def token() -> Optional[str]:
        """Gitlab token property

        :return: The Gitlab token environment variable (GL_TOKEN) value
        """
        return os.environ.get(config.get("gitlab_token_var"))

    @staticmethod
    @LoggedFunction(logger)
    def check_build_status(owner: str, repo: str, ref: str) -> bool:
        """Check last build status

        :param owner: The owner namespace of the repository. It includes all groups and subgroups.
        :param repo: The repository name
        :param ref: The sha1 hash of the commit ref

        :return: the status of the pipeline (False if a job failed)
        """
        gl = gitlab.Gitlab(Gitlab.api_url(), private_token=Gitlab.token())
        gl.auth()
        jobs = gl.projects.get(owner + "/" + repo).commits.get(ref).statuses.list()
        for job in jobs:
            if job["status"] not in ["success", "skipped"]:  # type: ignore[index]
                if job["status"] == "pending":  # type: ignore[index]
                    logger.debug(
                        f"check_build_status: job {job['name']} is still in pending status"  # type: ignore[index]
                    )
                    return False
                elif job["status"] == "failed" and not job["allow_failure"]:  # type: ignore[index]
                    logger.debug(f"check_build_status: job {job['name']} failed")  # type: ignore[index]
                    return False
        return True

    @classmethod
    @LoggedFunction(logger)
    def post_release_changelog(
        cls, owner: str, repo: str, version: str, changelog: str
    ) -> bool:
        """Post release changelog

        :param owner: The owner namespace of the repository
        :param repo: The repository name
        :param version: The version number
        :param changelog: The release notes for this version

        :return: The status of the request
        """
        ref = get_formatted_tag(version)
        gl = gitlab.Gitlab(Gitlab.api_url(), private_token=Gitlab.token())
        gl.auth()
        try:
            logger.debug(f"Before release create call")
            gl.projects.get(owner + "/" + repo).releases.create(
                {
                    "name": "Release " + version,
                    "tag_name": ref,
                    "description": changelog,
                }
            )
        except exceptions.GitlabCreateError:
            logger.debug(
                f"Release {ref} could not be created for project {owner}/{repo}"
            )
            return False

        return True


@LoggedFunction(logger)
def get_hvcs() -> Base:
    """Get HVCS helper class

    :raises ImproperConfigurationError: if the hvcs option provided is not valid
    """
    hvcs = config.get("hvcs")
    try:
        return globals()[hvcs.capitalize()]
    except KeyError:
        raise ImproperConfigurationError(
            '"{0}" is not a valid option for hvcs. Please use "github"|"gitlab"|"gitea"'.format(
                hvcs
            )
        )


def check_build_status(owner: str, repository: str, ref: str) -> bool:
    """
    Checks the build status of a commit on the api from your hosted version control provider.

    :param owner: The owner of the repository
    :param repository: The repository name
    :param ref: Commit or branch reference
    :return: A boolean with the build status
    """
    logger.debug(f"Checking build status for {owner}/{repository}#{ref}")
    return get_hvcs().check_build_status(owner, repository, ref)


def post_changelog(owner: str, repository: str, version: str, changelog: str) -> bool:
    """
    Posts the changelog to the current hvcs release API

    :param owner: The owner of the repository
    :param repository: The repository name
    :param version: A string with the new version
    :param changelog: A string with the changelog in correct format
    :return: a tuple with success status and payload from hvcs
    """
    logger.debug(f"Posting release changelog for {owner}/{repository} {version}")
    return get_hvcs().post_release_changelog(owner, repository, version, changelog)


def upload_to_release(owner: str, repository: str, version: str, path: str) -> bool:
    """
    Upload distributions to the current hvcs release API

    :param owner: The owner of the repository
    :param repository: The repository name
    :param version: A string with the version to upload for
    :param path: Path to dist directory

    :return: Status of the request
    """

    return get_hvcs().upload_dists(owner, repository, version, path)


def get_token() -> Optional[str]:
    """
    Returns the token for the current VCS

    :return: The token in string form
    """
    return get_hvcs().token()


def get_domain() -> Optional[str]:
    """
    Returns the domain for the current VCS

    :return: The domain in string form
    """
    return get_hvcs().domain()


def check_token() -> bool:
    """
    Checks whether there exists a token or not.

    :return: A boolean telling if there is a token.
    """
    return get_hvcs().token() is not None



"""CLI
"""
import logging
import os
import sys
from pathlib import Path

import click
import click_log

from semantic_release import ci_checks
from semantic_release.errors import GitError, ImproperConfigurationError

from .changelog import markdown_changelog
from .dist import build_dists, remove_dists, should_build, should_remove_dist
from .history import (
    evaluate_version_bump,
    get_current_version,
    get_new_version,
    get_previous_version,
    set_new_version,
)
from .history.logs import generate_changelog
from .hvcs import (
    check_build_status,
    check_token,
    get_domain,
    get_token,
    post_changelog,
    upload_to_release,
)
from .pre_commit import run_pre_commit, should_run_pre_commit
from .repository import ArtifactRepo
from .settings import config, overload_configuration
from .vcs_helpers import (
    checkout,
    commit_new_version,
    get_current_head_hash,
    get_repository_owner_and_name,
    push_new_version,
    tag_new_version,
    update_additional_files,
    update_changelog_file,
)

logger = logging.getLogger("semantic_release")

TOKEN_VARS = [
    "github_token_var",
    "gitlab_token_var",
    "pypi_pass_var",
    "pypi_token_var",
    "pypi_user_var",
    "repository_user_var",
    "repository_pass_var",
]

COMMON_OPTIONS = [
    click_log.simple_verbosity_option(logger),
    click.option(
        "--major", "force_level", flag_value="major", help="Force major version."
    ),
    click.option(
        "--minor", "force_level", flag_value="minor", help="Force minor version."
    ),
    click.option(
        "--patch", "force_level", flag_value="patch", help="Force patch version."
    ),
    click.option("--prerelease", is_flag=True, help="Creates a prerelease version."),
    click.option("--post", is_flag=True, help="Post changelog."),
    click.option("--retry", is_flag=True, help="Retry the same release, do not bump."),
    click.option(
        "--noop",
        is_flag=True,
        help="No-operations mode, finds the new version number without changing it.",
    ),
    click.option(
        "--define",
        "-D",
        multiple=True,
        help='setting="value", override a configuration value.',
    ),
    overload_configuration,
]


def common_options(func):
    """
    Decorator that adds all the options in COMMON_OPTIONS
    """
    for option in reversed(COMMON_OPTIONS):
        func = option(func)
    return func


def print_version(*, current=False, force_level=None, prerelease=False, **kwargs):
    """
    Print the current or new version to standard output.
    """
    try:
        current_version = get_current_version(prerelease)
    except GitError as e:
        print(str(e), file=sys.stderr)
        return False
    if current:
        print(current_version, end="")
        return True

    # Find what the new version number should be
    level_bump = evaluate_version_bump(current_version, force_level)
    new_version = get_new_version(current_version, level_bump, prerelease)
    if should_bump_version(current_version=current_version, new_version=new_version):
        print(new_version, end="")
        return True

    print("No release will be made.", file=sys.stderr)
    return False


def version(*, retry=False, noop=False, force_level=None, prerelease=False, **kwargs):
    """
    Detect the new version according to git log and semver.

    Write the new version number and commit it, unless the noop option is True.
    """
    if retry:
        logger.info("Retrying publication of the same version")
    else:
        logger.info("Creating new version")

    # Get the current version number
    try:
        current_version = get_current_version(prerelease)
        logger.info(f"Current version: {current_version}")
    except GitError as e:
        logger.error(str(e))
        return False
    # Find what the new version number should be
    level_bump = evaluate_version_bump(current_version, force_level)
    new_version = get_new_version(current_version, level_bump, prerelease)

    if not should_bump_version(
        current_version=current_version, new_version=new_version, retry=retry, noop=noop
    ):
        return False

    if retry:
        # No need to make changes to the repo, we're just retrying.
        return True

    # Bump the version
    bump_version(new_version, level_bump)
    return True


def should_bump_version(*, current_version, new_version, retry=False, noop=False):
    """Test whether the version should be bumped."""
    if new_version == current_version and not retry:
        logger.info("No release will be made.")
        return False

    if noop:
        logger.warning(
            "No operation mode. Should have bumped "
            f"from {current_version} to {new_version}"
        )
        return False

    if config.get("check_build_status"):
        logger.info("Checking build status...")
        owner, name = get_repository_owner_and_name()
        if not check_build_status(owner, name, get_current_head_hash()):
            logger.warning("The build failed, cancelling the release")
            return False
        logger.info("The build was a success, continuing the release")

    return True


def bump_version(new_version, level_bump):
    """
    Set the version to the given `new_version`.

    Edit in the source code, commit and create a git tag.
    """
    logger.info(f"Bumping with a {level_bump} version to {new_version}")
    if config.get("version_source") == "tag_only":
        tag_new_version(new_version)

        # we are done, no need for file changes if we are using
        # tags as version source
        return

    set_new_version(new_version)
    if config.get(
        "commit_version_number",
        config.get("version_source") == "commit",
    ):
        commit_new_version(new_version)
    if config.get("version_source") == "tag" or config.get("tag_commit"):
        tag_new_version(new_version)


def changelog(*, unreleased=False, noop=False, post=False, prerelease=False, **kwargs):
    """
    Generate the changelog since the last release.

    :raises ImproperConfigurationError: if there is no current version
    """
    current_version = get_current_version(prerelease)
    if current_version is None:
        raise ImproperConfigurationError(
            "Unable to get the current version. "
            "Make sure semantic_release.version_variable "
            "is setup correctly"
        )

    previous_version = get_previous_version(current_version)

    # Generate the changelog
    if unreleased:
        log = generate_changelog(current_version, None)
    else:
        log = generate_changelog(previous_version, current_version)

    owner, name = get_repository_owner_and_name()
    # print is used to keep the changelog on stdout, separate from log messages
    print(markdown_changelog(owner, name, current_version, log, header=False))

    # Post changelog to HVCS if enabled
    if not noop and post:
        if check_token():
            logger.info("Posting changelog to HVCS")
            post_changelog(
                owner,
                name,
                current_version,
                markdown_changelog(owner, name, current_version, log, header=False),
            )
        else:
            logger.error("Missing token: cannot post changelog to HVCS")


def publish(
    retry: bool = False, noop: bool = False, prerelease: bool = False, **kwargs
):
    """Run the version task, then push to git and upload to an artifact repository / GitHub Releases."""
    current_version = get_current_version(prerelease)

    verbose = logger.isEnabledFor(logging.DEBUG)
    if retry:
        logger.info("Retry is on")
        # The "new" version will actually be the current version, and the
        # "current" version will be the previous version.
        level_bump = None
        new_version = current_version
        current_version = get_previous_version(current_version)
    else:
        # Calculate the new version
        level_bump = evaluate_version_bump(current_version, kwargs.get("force_level"))
        new_version = get_new_version(current_version, level_bump, prerelease)

    owner, name = get_repository_owner_and_name()

    branch = config.get("branch")
    logger.debug(f"Running publish on branch {branch}")
    ci_checks.check(branch)
    checkout(branch)

    if should_bump_version(
        current_version=current_version,
        new_version=new_version,
        retry=retry,
        noop=noop,
    ):
        log = generate_changelog(current_version)
        changelog_md = markdown_changelog(
            owner,
            name,
            new_version,
            log,
            header=False,
            previous_version=current_version,
        )

        if should_run_pre_commit():
            logger.info("Running pre-commit command")
            run_pre_commit()

        if not retry:
            update_changelog_file(new_version, changelog_md)
            update_additional_files()
            bump_version(new_version, level_bump)
        # A new version was released
        logger.info("Pushing new version")
        push_new_version(
            auth_token=get_token(),
            owner=owner,
            name=name,
            branch=branch,
            domain=get_domain(),
        )

        # Get config options for uploads
        dist_path = config.get("dist_path")
        upload_release = config.get("upload_to_release")

        if should_build():
            # We need to run the command to build wheels for releasing
            logger.info("Building distributions")
            if should_remove_dist():
                # Remove old distributions before building
                remove_dists(dist_path)
            build_dists()

        if ArtifactRepo.upload_enabled():
            logger.info("Uploading to artifact Repository")
            ArtifactRepo(Path(dist_path)).upload(
                noop=noop, verbose=verbose, skip_existing=retry
            )

        if check_token():
            # Update changelog on HVCS
            logger.info("Posting changelog to HVCS")
            try:
                post_changelog(owner, name, new_version, changelog_md)
            except GitError:
                logger.error("Posting changelog failed")
        else:
            logger.warning("Missing token: cannot post changelog to HVCS")

        # Upload to GitHub Releases
        if upload_release:
            if check_token():
                logger.info("Uploading to HVCS release")
                upload_to_release(owner, name, new_version, dist_path)
                logger.info("Upload to HVCS is complete")
            else:
                logger.warning("Missing token: cannot upload to HVCS")

        # Remove distribution files as they are no longer needed
        if should_remove_dist():
            logger.info("Removing distribution files")
            remove_dists(dist_path)

        logger.info("Publish has finished")

    # else: Since version shows a message on failure, we do not need to print another.


def filter_output_for_secrets(message):
    """Remove secrets from cli output."""
    output = message
    for token_var in TOKEN_VARS:
        secret_name = config.get(token_var)
        secret = os.environ.get(secret_name)
        if secret != "" and secret is not None:
            output = output.replace(secret, f"${secret_name}")

    return output


def entry():
    # Move flags to after the command
    ARGS = sorted(sys.argv[1:], key=lambda x: 1 if x.startswith("--") else -1)

    if ARGS and not ARGS[0].startswith("print-"):
        # print-* command output should not be polluted with logging.
        click_log.basic_config()

    main(args=ARGS)


#
# Making the CLI commands.
# We have a level of indirection to the logical commands
# so we can successfully mock them during testing
#


@click.group()
@common_options
def main(**kwargs):
    logger.debug(f"Main args: {kwargs}")
    message = ""
    for token_var in TOKEN_VARS:
        secret_name = config.get(token_var)
        message += f'{secret_name}="{os.environ.get(secret_name)}",'
    logger.debug(f"Environment: {filter_output_for_secrets(message)}")

    obj = {}
    for key in [
        "check_build_status",
        "commit_subject",
        "commit_message",
        "commit_parser",
        "patch_without_tag",
        "major_on_zero",
        "upload_to_pypi",
        "upload_to_repository",
        "version_source",
        "no_git_tag",
    ]:
        obj[key] = config.get(key)
    logger.debug(f"Main config: {obj}")


@main.command(name="publish", help=publish.__doc__)
@common_options
def cmd_publish(**kwargs):
    try:
        return publish(**kwargs)
    except Exception as error:
        logger.error(filter_output_for_secrets(str(error)))
        exit(1)


@main.command(name="changelog", help=changelog.__doc__)
@common_options
@click.option(
    "--unreleased/--released",
    help="Decides whether to show the released or unreleased changelog.",
)
def cmd_changelog(**kwargs):
    try:
        return changelog(**kwargs)
    except Exception as error:
        logger.error(filter_output_for_secrets(str(error)))
        exit(1)


@main.command(name="version", help=version.__doc__)
@common_options
def cmd_version(**kwargs):
    try:
        return version(**kwargs)
    except Exception as error:
        logger.error(filter_output_for_secrets(str(error)))
        exit(1)


@main.command(name="print-version", help=print_version.__doc__)
@common_options
@click.option(
    "--current/--next",
    default=False,
    help="Choose to output next version (default) or current one.",
)
def cmd_print_version(**kwargs):
    try:
        return print_version(**kwargs)
    except Exception as error:
        print(filter_output_for_secrets(str(error)), file=sys.stderr)
        exit(1)



from typing import Optional

from ..settings import config
from ..vcs_helpers import get_formatted_tag, get_repository_owner_and_name


def get_github_compare_url(from_version: str, to_version: str) -> str:
    """
    Get the GitHub comparison link between two version tags.

    :param from_version: The older version to compare.
    :param to_version: The newer version to compare.
    :return: Link to view a comparison between the two versions.
    """
    owner, name = get_repository_owner_and_name()
    return (
        f"https://github.com/{owner}/{name}/compare/"
        f"{get_formatted_tag(from_version)}...{get_formatted_tag(to_version)}"
    )


def compare_url(version: str, previous_version: str = None, **kwargs) -> Optional[str]:
    if config.get("hvcs").lower() == "github" and previous_version:
        compare_url = get_github_compare_url(previous_version, version)
        return f"**[See all commits in this version]({compare_url})**"

    return None



"""Logs
"""
import logging
from typing import Optional

from ..errors import UnknownCommitMessageStyleError
from ..helpers import LoggedFunction
from ..settings import config, current_commit_parser
from ..vcs_helpers import get_commit_log

logger = logging.getLogger(__name__)

LEVELS = {
    0: None,
    1: "patch",
    2: "minor",
    3: "major",
}


@LoggedFunction(logger)
def evaluate_version_bump(current_version: str, force: str = None) -> Optional[str]:
    """
    Read git log since the last release to decide if we should make a major, minor or patch release.

    :param current_version: A string with the current version number.
    :param force: A string with the bump level that should be forced.
    :return: A string with either major, minor or patch if there should be a release.
             If no release is necessary, None will be returned.
    """
    if force:
        return force

    bump = None

    changes = []
    commit_count = 0

    for _hash, commit_message in get_commit_log(current_version):
        if commit_message.startswith(current_version):
            # Stop once we reach the current version
            # (we are looping in the order of newest -> oldest)
            logger.debug(
                f'"{commit_message}" is commit for {current_version}, breaking loop'
            )
            break

        commit_count += 1

        # Attempt to parse this commit using the currently-configured parser
        try:
            message = current_commit_parser()(commit_message)
            changes.append(message.bump)
        except UnknownCommitMessageStyleError as err:
            logger.debug(f"Ignoring UnknownCommitMessageStyleError: {err}")
            pass

    logger.debug(f"Commits found since last release: {commit_count}")

    if changes:
        # Select the largest required bump level from the commits we parsed
        level = max(changes)
        if level in LEVELS:
            bump = LEVELS[level]
        else:
            logger.warning(f"Unknown bump level {level}")

    if config.get("patch_without_tag") and commit_count > 0 and bump is None:
        bump = "patch"
        logger.debug(f"Changing bump level to patch based on config patch_without_tag")

    if (
        not config.get("major_on_zero")
        and current_version.startswith("0.")
        and bump == "major"
    ):
        bump = "minor"
        logger.debug("Changing bump level to minor based on config major_on_zero")

    return bump


@LoggedFunction(logger)
def generate_changelog(from_version: str, to_version: str = None) -> dict:
    """
    Parse a changelog dictionary for the given version.

    :param from_version: The version before where the changelog starts.
                         The changelog will be generated from the commit after this one.
    :param to_version: The last version included in the changelog.
    :return: A dict with changelog sections and commits
    """
    # Additional sections will be added as new types are encountered
    changes: dict = {"breaking": []}

    rev = None
    if from_version:
        rev = from_version

    found_the_release = to_version is None
    for _hash, commit_message in get_commit_log(rev):
        if not found_the_release:
            # Skip until we find the last commit in this release
            # (we are looping in the order of newest -> oldest)
            if to_version and to_version not in commit_message:
                continue
            else:
                logger.debug(
                    f"Reached the start of {to_version}, beginning changelog generation"
                )
                found_the_release = True

        if from_version is not None and from_version in commit_message:
            # We reached the previous release
            logger.debug(f"{from_version} reached, ending changelog generation")
            break

        try:
            message = current_commit_parser()(commit_message)
            if message.type not in changes:
                logger.debug(f"Creating new changelog section for {message.type} ")
                changes[message.type] = list()

            # Capitalize the first letter of the message, leaving others as they were
            # (using str.capitalize() would make the other letters lowercase)
            formatted_message = (
                message.descriptions[0][0].upper() + message.descriptions[0][1:]
            )
            if config.get("changelog_capitalize") is False:
                formatted_message = message.descriptions[0]

            # By default, feat(x): description shows up in changelog with the
            # scope bolded, like:
            #
            # * **x**: description
            if config.get("changelog_scope") and message.scope:
                formatted_message = f"**{message.scope}:** {formatted_message}"

            changes[message.type].append((_hash, formatted_message))

            if message.breaking_descriptions:
                # Copy breaking change descriptions into changelog
                for paragraph in message.breaking_descriptions:
                    changes["breaking"].append((_hash, paragraph))
            elif message.bump == 3:
                # Major, but no breaking descriptions, use commit subject instead
                changes["breaking"].append((_hash, message.descriptions[0]))

        except UnknownCommitMessageStyleError as err:
            logger.debug(f"Ignoring UnknownCommitMessageStyleError: {err}")

    return changes



import pytest

from semantic_release.errors import UnknownCommitMessageStyleError
from semantic_release.history import scipy_parser


def test_valid_scipy_commit(valid_scipy_commit, expected_response_scipy):
    (commit_tag, subject, _, body_parts) = expected_response_scipy
    result = scipy_parser(valid_scipy_commit)

    assert result[0] == commit_tag
    assert result[1] == subject
    assert len(result[3]) == len(body_parts)
    assert all(a == b for a, b in zip(result[3], body_parts))



"""VCS Helpers
"""
import logging
import os
import re
from datetime import date
from functools import wraps
from pathlib import Path, PurePath
from typing import List, Optional, Tuple
from urllib.parse import urlsplit

from git import GitCommandError, InvalidGitRepositoryError, Repo
from git.exc import BadName
from git.objects import TagObject

from .errors import GitError, HvcsRepoParseError
from .helpers import LoggedFunction
from .settings import config

try:
    repo = Repo(".", search_parent_directories=True)
except InvalidGitRepositoryError:
    pass

logger = logging.getLogger(__name__)


def check_repo(func):
    """Decorator which checks that we are in a git repository."""

    @wraps(func)
    def function_wrapper(*args, **kwargs):
        if repo is None:
            raise GitError("Not in a valid git repository")
        return func(*args, **kwargs)

    return function_wrapper


def get_formatted_tag(version):
    """Get the version, formatted with `tag_format` config option"""
    tag_format = config.get("tag_format")
    return tag_format.format(version=version)


@check_repo
def get_commit_log(from_rev=None):
    """Yield all commit messages from last to first."""
    rev = None
    if from_rev:
        from_rev = get_formatted_tag(from_rev)
        try:
            repo.commit(from_rev)
            rev = f"...{from_rev}"
        except BadName:
            logger.debug(
                f"Reference {from_rev} does not exist, considering entire history"
            )

    for commit in repo.iter_commits(rev):
        yield (commit.hexsha, commit.message.replace("\r\n", "\n"))


@check_repo
@LoggedFunction(logger)
def get_last_version(skip_tags=None, omit_pattern=None) -> Optional[str]:
    """
    Find the latest version using repo tags.

    :return: A string containing a version number.
    """
    skip_tags = skip_tags or []

    def version_finder(tag):
        if isinstance(tag.commit, TagObject):
            return tag.tag.tagged_date
        return tag.commit.committed_date

    for i in sorted(repo.tags, reverse=True, key=version_finder):
        match = re.search(r"\d+\.\d+\.\d+", i.name)
        if match:
            # check if the omit pattern is present in the tag (e.g. -beta for pre-release tags)
            if omit_pattern and omit_pattern in i.name:
                continue
            if i.name not in skip_tags:
                return match.group(0)  # Return only numeric vesion like 1.2.3

    return None


@check_repo
@LoggedFunction(logger)
def get_version_from_tag(tag_name: str) -> Optional[str]:
    """
    Get the git commit hash corresponding to a tag name.

    :param tag_name: Name of the git tag (i.e. 'v1.0.0')
    :return: sha1 hash of the commit
    """
    for i in repo.tags:
        if i.name == tag_name:
            return i.commit.hexsha
    return None


@check_repo
@LoggedFunction(logger)
def get_repository_owner_and_name() -> Tuple[str, str]:
    """
    Check the 'origin' remote to get the owner and name of the remote repository.

    :return: A tuple of the owner and name.
    """
    # Gitlab-CI context
    if "CI_PROJECT_NAMESPACE" in os.environ and "CI_PROJECT_NAME" in os.environ:
        return os.environ["CI_PROJECT_NAMESPACE"], os.environ["CI_PROJECT_NAME"]

    # Github actions context
    if "GITHUB_REPOSITORY" in os.environ:
        owner, name = os.environ["GITHUB_REPOSITORY"].rsplit("/", 1)
        return owner, name

    # Local context
    url = repo.remote("origin").url
    split_url = urlsplit(url)
    # Select the owner and name as regex groups
    parts = re.search(r"[:/]([^:]+)/([^/]*?)(.git)?$", split_url.path)
    if not parts:
        raise HvcsRepoParseError

    return parts.group(1), parts.group(2)


@check_repo
def get_current_head_hash() -> str:
    """
    Get the commit hash of the current HEAD.

    :return: The commit hash.
    """
    return repo.head.commit.name_rev.split(" ")[0]


@check_repo
@LoggedFunction(logger)
def commit_new_version(version: str):
    """
    Commit the file containing the version number variable.

    The commit message will be generated from the configured template.

    :param version: Version number to be used in the commit message.
    """
    from .history import load_version_declarations

    commit_subject = config.get("commit_subject")
    message = commit_subject.format(version=version)

    # Add an extended message if one is configured
    commit_message = config.get("commit_message")
    if commit_message:
        message += "\n\n"
        message += commit_message.format(version=version)

    commit_author = config.get(
        "commit_author",
        "semantic-release <semantic-release>",
    )

    for declaration in load_version_declarations():
        git_path: PurePath = PurePath(os.getcwd(), declaration.path).relative_to(repo.working_dir)  # type: ignore
        repo.git.add(str(git_path))

    return repo.git.commit(m=message, author=commit_author)


@check_repo
@LoggedFunction(logger)
def update_changelog_file(version: str, content_to_add: str):
    """
    Update changelog file with changelog for the release.

    :param version: The release version number, as a string.
    :param content_to_add: The release notes for the version.
    """
    changelog_file = config.get("changelog_file")
    changelog_placeholder = config.get("changelog_placeholder")
    git_path = Path(os.getcwd(), changelog_file)
    if not git_path.exists():
        original_content = f"# Changelog\n\n{changelog_placeholder}\n"
        logger.warning(f"Changelog file not found: {git_path} - creating it.")
    else:
        original_content = git_path.read_text()

    if changelog_placeholder not in original_content:
        logger.warning(
            f"Placeholder '{changelog_placeholder}' not found "
            f"in changelog file {git_path} - skipping change."
        )
        return

    updated_content = original_content.replace(
        changelog_placeholder,
        "\n".join(
            [
                changelog_placeholder,
                "",
                f"## v{version} ({date.today():%Y-%m-%d})",
                content_to_add,
            ]
        ),
    )
    git_path.write_text(updated_content)
    repo.git.add(str(git_path.relative_to(str(repo.working_dir))))


def get_changed_files(repo: Repo) -> List[str]:
    """
    Get untracked / dirty files in the given git repo.

    :param repo: Git repo to check.
    :return: A list of filenames.
    """
    untracked_files = repo.untracked_files
    dirty_files = [item.a_path for item in repo.index.diff(None)]
    return [*untracked_files, *dirty_files]


@check_repo
@LoggedFunction(logger)
def update_additional_files():
    """
    Add specified files to VCS, if they've changed.
    """
    changed_files = get_changed_files(repo)

    include_additional_files = config.get("include_additional_files")
    if include_additional_files:
        for filename in include_additional_files.split(","):
            if filename in changed_files:
                logger.debug(f"Updated file: {filename}")
                repo.git.add(filename)
            else:
                logger.warning(f"File {filename} shows no changes, cannot update it.")


@check_repo
@LoggedFunction(logger)
def tag_new_version(version: str):
    """
    Create a new tag with the version number, prefixed with v by default.

    :param version: The version number used in the tag as a string.
    """
    tag = get_formatted_tag(version)
    return repo.git.tag("-a", tag, m=tag)


@check_repo
@LoggedFunction(logger)
def push_new_version(
    auth_token: str = None,
    owner: str = None,
    name: str = None,
    branch: str = "master",
    domain: str = "github.com",
):
    """
    Run git push and git push --tags.

    :param auth_token: Authentication token used to push.
    :param owner: Organisation or user that owns the repository.
    :param name: Name of repository.
    :param branch: Branch to push to
    :param server_url: Name of the server. Will be used to identify a gitlab instance.
    :raises GitError: if GitCommandError is raised
    """
    server = "origin"
    if auth_token:
        if not config.get("ignore_token_for_push"):
            token = auth_token
            if config.get("hvcs") == "gitlab":
                token = "gitlab-ci-token:" + token
            actor = os.environ.get("GITHUB_ACTOR")
            if actor:
                server = f"http://{actor}:{token}@{domain}/{owner}/{name}.git"
            else:
                server = f"http://{token}@{domain}/{owner}/{name}.git"
        else:
            server = f"http://{token}@{domain}/{owner}/{name}.git"

    try:
        repo.git.push(server, branch)
        repo.git.push("--tags", server, branch)
    except GitCommandError as error:
        message = str(error)
        if auth_token:
            message = message.replace(auth_token, "[AUTH_TOKEN]")
        raise GitError(message)


@check_repo
@LoggedFunction(logger)
def checkout(branch: str):
    """
    Check out the given branch in the local repository.

    :param branch: The branch to checkout.
    """
    return repo.git.checkout(branch)



"""Semantic Release
"""
__version__ = "7.28.1"


from .errors import UnknownCommitMessageStyleError  # noqa; noqa
from .errors import ImproperConfigurationError, SemanticReleaseBaseError


def setup_hook(argv: list):
    """
    A hook to be used in setup.py to enable `python setup.py publish`.

    :param argv: sys.argv
    """
    if len(argv) > 1 and argv[1] in ["version", "publish", "changelog"]:
        from .cli import main

        main()



import gitlab

from .. import mock, wrapped_config_get

gitlab.Gitlab("")  # instantiation necessary to discover gitlab ProjectManager


class _GitlabProject:
    def __init__(self, status):
        self.commits = {"my_ref": self._Commit(status)}
        self.tags = self._Tags()
        self.releases = self._Releases()

    class _Commit:
        def __init__(self, status):
            self.statuses = self._Statuses(status)

        class _Statuses:
            def __init__(self, status):
                if status == "pending":
                    self.jobs = [
                        {
                            "name": "good_job",
                            "status": "passed",
                            "allow_failure": False,
                        },
                        {
                            "name": "slow_job",
                            "status": "pending",
                            "allow_failure": False,
                        },
                    ]
                elif status == "failure":
                    self.jobs = [
                        {
                            "name": "good_job",
                            "status": "passed",
                            "allow_failure": False,
                        },
                        {"name": "bad_job", "status": "failed", "allow_failure": False},
                    ]
                elif status == "allow_failure":
                    self.jobs = [
                        {
                            "name": "notsobad_job",
                            "status": "failed",
                            "allow_failure": True,
                        },
                        {
                            "name": "good_job2",
                            "status": "passed",
                            "allow_failure": False,
                        },
                    ]
                elif status == "success":
                    self.jobs = [
                        {
                            "name": "good_job1",
                            "status": "passed",
                            "allow_failure": True,
                        },
                        {
                            "name": "good_job2",
                            "status": "passed",
                            "allow_failure": False,
                        },
                    ]

            def list(self):
                return self.jobs

    class _Tags:
        def __init__(self):
            pass

        def get(self, version):
            if version == "vmy_good_tag":
                return self._Tag()
            elif version == "vmy_locked_tag":
                return self._Tag(locked=True)
            else:
                raise gitlab.exceptions.GitlabGetError

        class _Tag:
            def __init__(self, locked=False):
                self.locked = locked

            def set_release_description(self, changelog):
                if self.locked:
                    raise gitlab.exceptions.GitlabUpdateError

    class _Releases:
        def __init__(self):
            pass

        def create(self, input):
            if input["name"] and input["tag_name"]:
                if (
                    input["tag_name"] == "vmy_good_tag"
                    or input["tag_name"] == "vmy_locked_tag"
                ):
                    return self._Release()
            raise gitlab.exceptions.GitlabCreateError

        class _Release:
            def __init__(self, locked=False):
                pass


def mock_gitlab(status="success"):
    mocks = [
        mock.patch("os.environ", {"GL_TOKEN": "token"}),
        mock.patch(
            "semantic_release.hvcs.config.get", wrapped_config_get(hvcs="gitlab")
        ),
        mock.patch("gitlab.Gitlab.auth"),
        mock.patch(
            "gitlab.v4.objects.ProjectManager",
            return_value={"owner/repo": _GitlabProject(status)},
        ),
    ]

    def wraps(func):
        for option in reversed(mocks):
            func = option(func)
        return func

    return wraps



import pytest

from semantic_release import ci_checks
from semantic_release.errors import CiVerificationError


def test_gitlab_should_pass_if_branch_is_master(monkeypatch):
    monkeypatch.setenv("CI_COMMIT_REF_NAME", "master")

    assert ci_checks.gitlab("master")


def test_gitlab_should_pass_if_branch_is_correct(monkeypatch):
    monkeypatch.setenv("CI_COMMIT_REF_NAME", "other-branch")

    assert ci_checks.gitlab("other-branch")


def test_gitlab_should_raise_ci_verification_error_for_wrong_branch(monkeypatch):
    monkeypatch.setenv("CI_COMMIT_REF_NAME", "other-branch")

    with pytest.raises(CiVerificationError):
        ci_checks.gitlab("master")



import pytest

from semantic_release.history.parser_scipy import COMMIT_TYPES, ChangeType


@pytest.fixture(params=COMMIT_TYPES)
def scipy_type(request):
    return request.param


@pytest.fixture(
    params=[
        "scipy.stats.qmc: centered-discrepancy optimization of a Latin hypercube",
        "inverse missing in idstn, idctn (#14479)",
        "Merge pull request #14447 from AnirudhDagar/rename_ndimage_modules",
        "Add tests for args kwarg in quad_vec",
        "badge with version of the doc in the navbar (#14132)",
        "Bump scipy from 1.7.0 to 1.7.1 (#28)",
        "avoid nan if angle=0 in RotvecRotation",
    ]
)
def subject(request):
    return request.param


@pytest.fixture(
    params=[
        # a squash merge that preserved PR commit messages
        (
            "DOC: import ropy.transform to test for numpy error",
            "DOC: lower numpy version",
            "DOC: lower numpy version further",
            "MAINT: remove debugging import",
        ),
        # empty body
        list(),
        # formatted body
        (
            """Bumps [sphinx](https://github.com/sphinx-doc/sphinx) from 3.5.3 to 4.1.1.
            - [Release notes](https://github.com/sphinx-doc/sphinx/releases)
            - [Changelog](https://github.com/sphinx-doc/sphinx/blob/4.x/CHANGES)
            - [Commits](https://github.com/sphinx-doc/sphinx/commits/v4.1.1)""",
            """---
            updated-dependencies:
            - dependency-name: sphinx
            dependency-type: direct:development
            update-type: version-update:semver-major""",
        ),
        (
            "Bug spotted on Fedora, see https://src.fedoraproject.org/rpms/scipy/pull-request/22",
            "The `int[::]` annotation is used to accept non-contiguous views.",
        ),
        ("[skip azp] [skip actions]",),
    ]
)
def body_parts(request):
    return request.param


@pytest.fixture()
def expected_response_scipy(scipy_type: ChangeType, subject, body_parts):
    bump_level = scipy_type.bump_level
    topic = scipy_type.section
    changelog_body = (subject, *body_parts)
    return (bump_level, topic, None, changelog_body)


@pytest.fixture()
def valid_scipy_commit(scipy_type: ChangeType, subject, body_parts):
    body = "\n\n".join(body_parts)
    commit_msg = f"{scipy_type.tag}: {subject}\n\n{body}"
    return commit_msg



import pytest

from semantic_release import ci_checks
from semantic_release.errors import CiVerificationError


def test_jenkins_should_pass_if_url_set_branch_master_no_pr(monkeypatch):
    monkeypatch.setenv("JENKINS_URL", "http://custom.jenkins.int.com")
    monkeypatch.setenv("BRANCH_NAME", "master")

    assert ci_checks.jenkins("master")


def test_jenkins_should_pass_if_url_set_if_branch_correct_no_pr(monkeypatch):
    monkeypatch.setenv("JENKINS_URL", "http://custom.jenkins.int.com")
    monkeypatch.setenv("GIT_BRANCH", "foo-bar")

    assert ci_checks.jenkins("foo-bar")


def test_jenkins_should_raise_ci_verification_error_no_url(monkeypatch):
    monkeypatch.setenv("BRANCH_NAME", "foo-bar")

    with pytest.raises(CiVerificationError):
        assert ci_checks.jenkins("foo-bar")


def test_jenkins_should_raise_ci_verification_error_for_pr(monkeypatch):
    monkeypatch.setenv("JENKINS_URL", "http://custom.jenkins.int.com")
    monkeypatch.setenv("BRANCH_NAME", "foo-bar")
    monkeypatch.setenv("CHANGE_ID", "1")

    with pytest.raises(CiVerificationError):
        assert ci_checks.jenkins("foo-bar")



# -*- coding: utf-8 -*-
"""Custom Errors
"""


class SemanticReleaseBaseError(Exception):
    pass


class ImproperConfigurationError(SemanticReleaseBaseError):
    pass


class UnknownCommitMessageStyleError(SemanticReleaseBaseError):
    pass


class GitError(SemanticReleaseBaseError):
    pass


class CiVerificationError(SemanticReleaseBaseError):
    pass


class HvcsRepoParseError(SemanticReleaseBaseError):
    pass



import functools
from typing import Union

from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


def format_arg(value):
    if type(value) == str:
        return f"'{value.strip()}'"
    else:
        return str(value)


def build_requests_session(
    raise_for_status=True, retry: Union[bool, int, Retry] = True
) -> Session:
    """
    Create a requests session.
    :param raise_for_status: If True, a hook to invoke raise_for_status be installed
    :param retry: If true, it will use default Retry configuration. if an integer, it will use default Retry
    configuration with given integer as total retry count. if Retry instance, it will use this instance.
    :return: configured requests Session
    """
    session = Session()
    if raise_for_status:
        session.hooks = {"response": [lambda r, *args, **kwargs: r.raise_for_status()]}
    if retry:
        if isinstance(retry, bool):
            retry = Retry()
        elif isinstance(retry, int):
            retry = Retry(retry)
        elif not isinstance(retry, Retry):
            raise ValueError("retry should be a bool, int or Retry instance.")
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("http://", adapter)
        session.mount("https://", adapter)
    return session


class LoggedFunction:
    """
    Decorator which adds debug logging to a function.

    The input arguments are logged before the function is called, and the
    return value is logged once it has completed.

    :param logger: Logger to send output to.
    """

    def __init__(self, logger):
        self.logger = logger

    def __call__(self, func):
        @functools.wraps(func)
        def logged_func(*args, **kwargs):
            # Log function name and arguments
            self.logger.debug(
                "{function}({args}{kwargs})".format(
                    function=func.__name__,
                    args=", ".join([format_arg(x) for x in args]),
                    kwargs="".join(
                        [f", {k}={format_arg(v)}" for k, v in kwargs.items()]
                    ),
                )
            )

            # Call function
            result = func(*args, **kwargs)

            # Log result
            if result is not None:
                self.logger.debug(f"{func.__name__} -> {result}")
            return result

        return logged_func



"""Helpers to read settings from setup.cfg or pyproject.toml
"""
import configparser
import importlib
import logging
import os
from collections import UserDict
from functools import wraps
from os import getcwd
from typing import Callable, List

import tomlkit
from tomlkit.exceptions import TOMLKitError

from .errors import ImproperConfigurationError

logger = logging.getLogger(__name__)


def _config():
    cwd = getcwd()
    ini_paths = [
        os.path.join(os.path.dirname(__file__), "defaults.cfg"),
        os.path.join(cwd, "setup.cfg"),
    ]
    ini_config = _config_from_ini(ini_paths)

    toml_path = os.path.join(cwd, "pyproject.toml")
    toml_config = _config_from_pyproject(toml_path)

    # Cast to a UserDict so that we can mock the get() method.
    return UserDict({**ini_config, **toml_config})


def _config_from_ini(paths):
    parser = configparser.ConfigParser()
    parser.read(paths)

    flags = {
        "changelog_capitalize",
        "changelog_scope",
        "check_build_status",
        "commit_version_number",
        "patch_without_tag",
        "major_on_zero",
        "remove_dist",
        "upload_to_pypi",
        "upload_to_repository",
        "upload_to_release",
        "tag_commit",
    }

    # Iterate through the sections so that default values are applied
    # correctly.  See:
    # https://stackoverflow.com/questions/1773793/convert-configparser-items-to-dictionary
    config = {}
    for key, _ in parser.items("semantic_release"):
        if key in flags:
            config[key] = parser.getboolean("semantic_release", key)
        else:
            config[key] = parser.get("semantic_release", key)

    return config


def _config_from_pyproject(path):
    if os.path.isfile(path):
        try:
            with open(path, "r") as f:
                pyproject = tomlkit.loads(f.read())
            if pyproject:
                return pyproject.get("tool", {}).get("semantic_release", {})
        except TOMLKitError as e:
            logger.warning(f"Could not decode pyproject.toml: {e}")

    return {}


config = _config()


def current_commit_parser() -> Callable:
    """Get the currently-configured commit parser

    :raises ImproperConfigurationError: if ImportError or AttributeError is raised
    :returns: Commit parser
    """

    try:
        # All except the last part is the import path
        parts = config.get("commit_parser").split(".")
        module = ".".join(parts[:-1])
        # The final part is the name of the parse function
        return getattr(importlib.import_module(module), parts[-1])
    except (ImportError, AttributeError) as error:
        raise ImproperConfigurationError(f'Unable to import parser "{error}"')


def current_changelog_components() -> List[Callable]:
    """Get the currently-configured changelog components

    :raises ImproperConfigurationError: if ImportError or AttributeError is raised
    :returns: List of component functions
    """
    component_paths = config.get("changelog_components").split(",")
    components = list()

    for path in component_paths:
        try:
            # All except the last part is the import path
            parts = path.split(".")
            module = ".".join(parts[:-1])
            # The final part is the name of the component function
            components.append(getattr(importlib.import_module(module), parts[-1]))
        except (ImportError, AttributeError) as error:
            raise ImproperConfigurationError(
                f'Unable to import changelog component "{path}"'
            )

    return components


def overload_configuration(func):
    """This decorator gets the content of the "define" array and edits "config"
    according to the pairs of key/value.
    """

    @wraps(func)
    def wrap(*args, **kwargs):
        if "define" in kwargs:
            for defined_param in kwargs["define"]:
                pair = defined_param.split("=", maxsplit=1)
                if len(pair) == 2:
                    config[str(pair[0])] = pair[1]
        return func(*args, **kwargs)

    return wrap



import re
import sys

from setuptools import find_packages, setup


def _read_long_description():
    try:
        with open("readme.rst") as fd:
            return fd.read()
    except Exception:
        return None


with open("semantic_release/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="python-semantic-release-viettel",
    version="1.0.0",
    url="http://github.com/relekang/python-semantic-release",
    author="congdd4",
    author_email="congdd4@viettel.com.vn",
    description="Custom Automatic Semantic Versioning for Python projects",
    long_description=_read_long_description(),
    packages=find_packages(exclude=("tests",)),
    license="MIT",
    install_requires=[
        "click>=7,<9",
        "click_log>=0.3,<1",
        "gitpython>=3.0.8,<4",
        "invoke>=1.4.1,<2",
        "semver>=2.10,<3",
        "twine>=3,<4",
        "requests>=2.25,<3",
        "wheel",
        "python-gitlab>=2,<4",
        # tomlkit used to be pinned to 0.7.0
        # See https://github.com/relekang/python-semantic-release/issues/336
        # and https://github.com/relekang/python-semantic-release/pull/337
        "tomlkit>=0.10.0,<0.11.0",
        "dotty-dict>=1.3.0,<2",
        "dataclasses==0.8; python_version < '3.7.0'",
    ],
    extras_require={
        "test": [
            "coverage>=5,<6",
            "pytest>=5,<6",
            "pytest-xdist>=1,<2",
            "pytest-mock>=2,<3",
            "responses==0.13.3",
            "mock==1.3.0",
        ],
        "docs": ["Sphinx==1.3.6"],
        "dev": ["tox", "isort", "black"],
        "mypy": ["mypy", "types-requests"],
    },
    entry_points="""
        [console_scripts]
        semantic-release=semantic_release.cli:entry
    """,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



"""Run commands prior to the release commit
"""
import logging

from invoke import run

from .settings import config

logger = logging.getLogger(__name__)


def should_run_pre_commit():
    command = config.get("pre_commit_command")
    return bool(command)


def run_pre_commit():
    command = config.get("pre_commit_command")
    logger.debug(f"Running {command}")
    run(command)



"""Commit parser helpers
"""
import collections
import re
from typing import List

re_breaking = re.compile("BREAKING[ -]CHANGE: ?(.*)")


ParsedCommit = collections.namedtuple(
    "ParsedCommit", ["bump", "type", "scope", "descriptions", "breaking_descriptions"]
)


def parse_paragraphs(text: str) -> List[str]:
    """
    This will take a text block and return a tuple containing each
    paragraph with single line breaks collapsed into spaces.

    :param text: The text string to be divided.
    :return: A tuple of paragraphs.
    """
    return [
        paragraph.replace("\n", " ")
        for paragraph in text.split("\n\n")
        if len(paragraph) > 0
    ]



import pytest

from semantic_release.errors import UnknownCommitMessageStyleError
from semantic_release.history import angular_parser

from .. import mock, wrapped_config_get

text = (
    "This is an long explanatory part of a commit message. It should give "
    "some insight to the fix this commit adds to the codebase."
)
footer = "Closes #400"


def test_parser_raises_unknown_message_style():
    pytest.raises(UnknownCommitMessageStyleError, angular_parser, "")
    pytest.raises(
        UnknownCommitMessageStyleError,
        angular_parser,
        "feat(parser\n): Add new parser pattern",
    )


def test_parser_return_correct_bump_level():
    assert (
        angular_parser("feat(parsers): Add new parser pattern\n\nBREAKING CHANGE: ")[0]
        == 3
    )
    assert (
        angular_parser("feat(parsers)!: Add new parser pattern\n\nBREAKING CHANGE: ")[0]
        == 3
    )
    assert (
        angular_parser(
            "feat(parsers): Add new parser pattern\n\n"
            "New pattern is awesome\n\nBREAKING CHANGE: "
        )[0]
        == 3
    )
    assert (
        angular_parser(
            "feat(parsers): Add new parser pattern\n\nBREAKING-CHANGE: change !"
        )[0]
        == 3
    )
    assert angular_parser("feat(parser): Add emoji parser")[0] == 2
    assert angular_parser("fix(parser): Fix regex in angular parser")[0] == 1
    assert angular_parser("test(parser): Add a test for angular parser")[0] == 0
    assert angular_parser("feat(parser)!: Edit dat parsing stuff")[0] == 3
    assert angular_parser("fix!: Edit dat parsing stuff again")[0] == 3


def test_parser_return_type_from_commit_message():
    assert angular_parser("feat(parser): ...")[1] == "feature"
    assert angular_parser("fix(parser): ...")[1] == "fix"
    assert angular_parser("test(parser): ...")[1] == "test"
    assert angular_parser("docs(parser): ...")[1] == "documentation"
    assert angular_parser("style(parser): ...")[1] == "style"
    assert angular_parser("refactor(parser): ...")[1] == "refactor"
    assert angular_parser("chore(parser): ...")[1] == "chore"


def test_parser_return_scope_from_commit_message():
    assert angular_parser("chore(parser): ...")[2] == "parser"
    assert angular_parser("chore(a part): ...")[2] == "a part"
    assert angular_parser("chore(a_part): ...")[2] == "a_part"
    assert angular_parser("chore(a-part): ...")[2] == "a-part"
    assert angular_parser("chore(a.part): ...")[2] == "a.part"
    assert angular_parser("chore(a+part): ...")[2] == "a+part"
    assert angular_parser("chore(a&part): ...")[2] == "a&part"
    assert angular_parser("chore((part)): ...")[2] == "(part)"
    assert angular_parser("chore((p):rt): ...")[2] == "(p):rt"


def test_parser_return_subject_from_commit_message():
    assert angular_parser("feat(parser): Add emoji parser")[3][0] == "Add emoji parser"
    assert (
        angular_parser("fix(parser): Fix regex in angular parser")[3][0]
        == "Fix regex in angular parser"
    )
    assert (
        angular_parser("test(parser): Add a test for angular parser")[3][0]
        == "Add a test for angular parser"
    )


def test_parser_return_text_from_commit_message():
    assert (
        angular_parser("fix(parser): Fix regex in an parser\n\n{}".format(text))[3][1]
        == text
    )


def test_parser_return_footer_from_commit_message():
    commit = "fix(tox): Fix env \n\n{t[text]}\n\n{t[footer]}".format(t=globals())
    assert angular_parser(commit)[3][2] == footer


def test_parser_should_accept_message_without_scope():
    assert angular_parser("fix: superfix")[0] == 1
    assert angular_parser("fix: superfix")[3][0] == "superfix"


##############################
# test custom parser options #
##############################
@mock.patch(
    "semantic_release.history.parser_angular.config.get",
    wrapped_config_get(parser_angular_default_level_bump="minor"),
)
def test_parser_custom_default_level():
    assert angular_parser("test(parser): Add a test for angular parser")[0] == 2


@mock.patch(
    "semantic_release.history.parser_angular.config.get",
    wrapped_config_get(
        parser_angular_allowed_types="custom,build,chore,ci,docs,fix,perf,style,refactor,test"
    ),
)
def test_parser_custom_allowed_types():
    assert angular_parser("custom: ...")[0] == 0
    assert angular_parser("custom(parser): ...")[1] == "custom"
    pytest.raises(UnknownCommitMessageStyleError, angular_parser, "feat(parser): ...")


@mock.patch(
    "semantic_release.history.parser_angular.config.get",
    wrapped_config_get(parser_angular_minor_types="docs"),
)
def test_parser_custom_minor_types():
    assert angular_parser("docs: write some docs")[0] == 2


@mock.patch(
    "semantic_release.history.parser_angular.config.get",
    wrapped_config_get(parser_angular_patch_types="test"),
)
def test_parser_custom_patch_types():
    assert angular_parser("test(this): added a test")[0] == 1



import re
from typing import Iterable, Optional

from ..hvcs import Github, Gitlab
from ..settings import config


def add_pr_link(owner: str, repo_name: str, message: str) -> str:
    """
    GitHub release notes automagically link to the PR, but changelog markdown
    doesn't. Replace (#123) at the end of a message with a markdown link.
    """

    pr_pattern = re.compile(r"\s+\(#(\d{1,8})\)$")
    match = re.search(pr_pattern, message)

    if match:
        pr_number = match.group(1)
        url = (
            f"http://{Gitlab.domain()}/{owner}/{repo_name}/-/issues/{pr_number}"
            if config.get("hvcs") == "gitlab"
            else f"http://{Github.domain()}/{owner}/{repo_name}/issues/{pr_number}"
        )

        return re.sub(pr_pattern, f" ([#{pr_number}]({url}))", message)

    return message


def get_changelog_sections(changelog: dict, changelog_sections: list) -> Iterable[str]:
    """Generator which yields each changelog section to be included"""

    included_sections = config.get("changelog_sections")
    included_sections = [s.strip() for s in included_sections.split(",")]

    for section in included_sections:
        if section in changelog and changelog[section]:
            yield section


def get_hash_link(owner: str, repo_name: str, hash_: str) -> str:
    """Generate the link for commit hash"""
    url = (
        f"http://{Gitlab.domain()}/{owner}/{repo_name}/-/commit/{hash_}"
        if config.get("hvcs") == "gitlab"
        else f"http://{Github.domain()}/{owner}/{repo_name}/commit/{hash_}"
    )
    short_hash = hash_[:7]
    return f"[`{short_hash}`]({url})"


def changelog_headers(
    owner: str, repo_name: str, changelog: dict, changelog_sections: list, **kwargs
) -> Optional[str]:
    output = ""

    for section in get_changelog_sections(changelog, changelog_sections):
        # Add a header for this section
        output += f"\n### {section.capitalize()}\n"

        # Add each commit from the section in an unordered list
        for item in changelog[section]:
            message = add_pr_link(owner, repo_name, item[1])
            output += f"* {message} ({get_hash_link(owner, repo_name, item[0])})\n"

    return output


def changelog_table(
    owner: str, repo_name: str, changelog: dict, changelog_sections: list, **kwargs
) -> str:
    output = "| Type | Change |\n| --- | --- |\n"

    for section in get_changelog_sections(changelog, changelog_sections):
        items = "<br>".join(
            [
                f"{add_pr_link(owner, repo_name, item[1])} "
                f"({get_hash_link(owner, repo_name, item[0])})"
                for item in changelog[section]
            ]
        )
        output += f"| {section.title()} | {items} |\n"

    return output



"""Build and manage distributions
"""
import logging
import shutil

from invoke import run

from .settings import config

logger = logging.getLogger(__name__)


def should_build():
    upload_to_artifact_repository = config.get("upload_to_repository") and config.get(
        "upload_to_pypi"
    )
    upload_release = config.get("upload_to_release")
    build_command = config.get("build_command")
    build_command = build_command if build_command != "false" else False
    return bool(build_command and (upload_to_artifact_repository or upload_release))


def should_remove_dist():
    remove_dist = config.get("remove_dist")
    return bool(remove_dist and should_build())


def build_dists():
    command = config.get("build_command")
    logger.info(f"Running {command}")
    run(command)


def remove_dists(path: str):
    logger.debug(f"Removing build folder: `{path}`")
    shutil.rmtree(path, ignore_errors=True)



"""Legacy commit parser from Python Semantic Release 1.0"""
import logging
import re
from typing import Optional

from ..errors import UnknownCommitMessageStyleError
from ..helpers import LoggedFunction
from ..settings import config
from .parser_helpers import ParsedCommit, parse_paragraphs, re_breaking

logger = logging.getLogger(__name__)

re_parser = re.compile(r"(?P<subject>[^\n]+)" r"(:?\n\n(?P<text>.+))?", re.DOTALL)


@LoggedFunction(logger)
def parse_commit_message(
    message: str,
) -> ParsedCommit:
    """
    Parse a commit message according to the 1.0 version of python-semantic-release.

    It expects a tag of some sort in the commit message and will use the rest of the first line
    as changelog content.

    :param message: A string of a commit message.
    :raises UnknownCommitMessageStyleError: If it does not recognise the commit style
    :return: A tuple of (level to bump, type of change, scope of change, a tuple with descriptions)
    """

    # Attempt to parse the commit message with a regular expression
    parsed = re_parser.match(message)
    if not parsed:
        raise UnknownCommitMessageStyleError(
            f"Unable to parse the given commit message: {message}"
        )

    subject = parsed.group("subject")

    # Check tags for minor or patch
    if config.get("minor_tag") in message:
        level = "feature"
        level_bump = 2
        if subject:
            subject = subject.replace(config.get("minor_tag"), "")

    elif config.get("fix_tag") in message:
        level = "fix"
        level_bump = 1
        if subject:
            subject = subject.replace(config.get("fix_tag"), "")

    else:
        # We did not find any tags in the commit message
        raise UnknownCommitMessageStyleError(
            f"Unable to parse the given commit message: {message}"
        )

    if parsed.group("text"):
        descriptions = parse_paragraphs(parsed.group("text"))
    else:
        descriptions = list()
    descriptions.insert(0, subject.strip())

    # Look for descriptions of breaking changes
    breaking_descriptions = [
        match.group(1)
        for match in (re_breaking.match(p) for p in descriptions[1:])
        if match
    ]
    if breaking_descriptions:
        level = "breaking"
        level_bump = 3

    return ParsedCommit(level_bump, level, None, descriptions, breaking_descriptions)



"""
Angular commit style parser

https://github.com/angular/angular/blob/master/CONTRIBUTING.md#-commit-message-guidelines
"""
import logging
import re

from ..errors import ImproperConfigurationError, UnknownCommitMessageStyleError
from ..helpers import LoggedFunction
from ..settings import config
from .parser_helpers import ParsedCommit, parse_paragraphs, re_breaking

logger = logging.getLogger(__name__)


# types with long names in changelog
LONG_TYPE_NAMES = {
    "feat": "feature",
    "docs": "documentation",
    "perf": "performance",
}

LEVEL_BUMPS = {"no-release": 0, "patch": 1, "minor": 2, "major": 3}


@LoggedFunction(logger)
def parse_commit_message(message: str) -> ParsedCommit:
    """
    Parse a commit message according to the angular commit guidelines specification.

    :param message: A string of a commit message.
    :return: A tuple of (level to bump, type of change, scope of change, a tuple with descriptions)
    :raises UnknownCommitMessageStyleError: if regular expression matching fails
    """

    # loading these are here to make it easier to mock in tests
    allowed_types = config.get("parser_angular_allowed_types").split(",")
    minor_types = config.get("parser_angular_minor_types").split(",")
    patch_types = config.get("parser_angular_patch_types").split(",")
    re_parser = re.compile(
        r"(?P<type>" + "|".join(allowed_types) + ")"
        r"(?:\((?P<scope>[^\n]+)\))?"
        r"(?P<break>!)?: "
        r"(?P<subject>[^\n]+)"
        r"(:?\n\n(?P<text>.+))?",
        re.DOTALL,
    )

    # Attempt to parse the commit message with a regular expression
    parsed = re_parser.match(message)
    if not parsed:
        raise UnknownCommitMessageStyleError(
            f"Unable to parse the given commit message: {message}"
        )
    parsed_break = parsed.group("break")
    parsed_scope = parsed.group("scope")
    parsed_subject = parsed.group("subject")
    parsed_text = parsed.group("text")
    parsed_type = parsed.group("type")

    if parsed_text:
        descriptions = parse_paragraphs(parsed_text)
    else:
        descriptions = list()
    # Insert the subject before the other paragraphs
    descriptions.insert(0, parsed_subject)

    # Look for descriptions of breaking changes
    breaking_descriptions = [
        match.group(1)
        for match in (re_breaking.match(p) for p in descriptions[1:])
        if match
    ]

    default_level_bump = config.get("parser_angular_default_level_bump").lower()
    if default_level_bump not in LEVEL_BUMPS.keys():
        raise ImproperConfigurationError(
            f"{default_level_bump} is not a valid option for "
            f"parser_angular_default_level_bump.\n"
            f"valid options are: {', '.join(LEVEL_BUMPS.keys())}"
        )
    level_bump = LEVEL_BUMPS[default_level_bump]
    if parsed_break or breaking_descriptions:
        level_bump = 3  # Major
    elif parsed_type in minor_types:
        level_bump = 2  # Minor
    elif parsed_type in patch_types:
        level_bump = 1  # Patch

    parsed_type_long = LONG_TYPE_NAMES.get(parsed_type, parsed_type)
    # first param is the key you're getting from the dict,
    # second param is the default value
    # allows only putting types with a long name in the LONG_TYPE_NAMES dict

    return ParsedCommit(
        level_bump,
        parsed_type_long,
        parsed_scope,
        descriptions,
        breaking_descriptions,
    )



"""
Parses commit messages using `scipy tags <scipy-style>`_ of the form::

    <tag>(<scope>): <subject>

    <body>


The elements <tag>, <scope> and <body> are optional. If no tag is present, the
commit will be added to the changelog section "None" and no version increment
will be performed.

While <scope> is supported here it isn't actually part of the scipy style.
If it is missing, parentheses around it are too. The commit should then be
of the form::

    <tag>: <subject>

    <body>

To communicate a breaking change add "BREAKING CHANGE" into the body at the
beginning of a paragraph. Fill this paragraph with information how to migrate
from the broken behavior to the new behavior. It will be added to the
"Breaking" section of the changelog.

Supported Tags::

    API, DEP, ENH, REV, BUG, MAINT, BENCH, BLD,
    DEV, DOC, STY, TST, REL, FEAT, TEST

Supported Changelog Sections::

    breaking, feature, fix, Other, None

.. _`scipy-style`: https://docs.scipy.org/doc/scipy/reference/dev/contributor/development_workflow.html#writing-the-commit-message
"""

import logging
import re

from ..errors import UnknownCommitMessageStyleError
from ..helpers import LoggedFunction
from .parser_helpers import ParsedCommit

logger = logging.getLogger(__name__)


class ChangeType:
    def __init__(self, tag, section) -> None:
        self.tag: str = tag
        self.section: str = section
        self.bump_level: int = 0

    def make_breaking(self):
        self.bump_level = 3


class Breaking(ChangeType):
    def __init__(self, tag, section) -> None:
        super().__init__(tag, section)
        self.bump_level: int = 3


class Compatible(ChangeType):
    def __init__(self, tag, section) -> None:
        super().__init__(tag, section)
        self.bump_level: int = 2


class Patch(ChangeType):
    def __init__(self, tag, section) -> None:
        super().__init__(tag, section)
        self.bump_level: int = 1


class Ignore(ChangeType):
    def __init__(self, tag, section) -> None:
        super().__init__(tag, section)
        self.bump_level: int = 0


COMMIT_TYPES = [
    Breaking("API", "breaking"),
    Ignore("BENCH", "None"),
    Patch("BLD", "fix"),
    Patch("BUG", "fix"),
    Compatible("DEP", "breaking"),
    Compatible("DEV", "None"),
    Ignore("DOC", "documentation"),
    Compatible("ENH", "feature"),
    Patch("MAINT", "fix"),
    Compatible("REV", "Other"),
    Ignore("STY", "None"),
    Ignore("TST", "None"),
    Ignore("REL", "None"),
    # strictly speaking not part of the standard
    Compatible("FEAT", "feature"),
    Ignore("TEST", "None"),
]

_commit_filter = "|".join(c.tag for c in COMMIT_TYPES)
re_parser = re.compile(
    rf"(?P<tag>{_commit_filter})?"
    r"(?:\((?P<scope>[^\n]+)\))?"
    r":? "
    r"(?P<subject>[^\n]+):?"
    r"(\n\n(?P<text>.*))?",
    re.DOTALL,
)


@LoggedFunction(logger)
def parse_commit_message(message: str) -> ParsedCommit:
    """
    Parse a scipy-style commit message

    :param message: A string of a commit message.
    :return: A tuple of (level to bump, type of change, scope of change, a tuple
    with descriptions)
    :raises UnknownCommitMessageStyleError: if regular expression matching fails
    """

    parsed = re_parser.match(message)
    if not parsed:
        raise UnknownCommitMessageStyleError(
            f"Unable to parse the given commit message: {message}"
        )

    if parsed.group("subject"):
        subject = parsed.group("subject")
    else:
        raise UnknownCommitMessageStyleError(f"The commit has no subject {message}")

    if parsed.group("text"):
        blocks = parsed.group("text").split("\n\n")
        blocks = [x for x in blocks if not x == ""]
        blocks.insert(0, subject)
    else:
        blocks = [subject]

    msg_type: ChangeType
    for msg_type in COMMIT_TYPES:
        if msg_type.tag == parsed.group("tag"):
            break
    else:
        # some commits may not have a tag, e.g. if they belong to a PR that
        # wasn't squashed (for maintainability) ignore them
        msg_type = Ignore("", "None")

    # Look for descriptions of breaking changes
    migration_instructions = [
        block for block in blocks if block.startswith("BREAKING CHANGE")
    ]
    if migration_instructions:
        msg_type.make_breaking()

    return ParsedCommit(
        msg_type.bump_level,
        msg_type.section,
        parsed.group("scope"),
        blocks,
        migration_instructions,
    )
